import json
from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from tests import aws_verified
from tests.test_stepfunctions.parser import sfn_role_policy
from tests.test_stepfunctions.test_stepfunctions import simple_definition


@pytest.mark.parametrize("use_parser", [True, False], ids=["use_parser", "use_mock"])
def test_describe_state_machine_using_version_arn(use_parser):
    with mock_aws(config={"stepfunctions": {"execute_state_machine": use_parser}}):
        iam = boto3.client("iam", region_name="us-east-1")
        role_name = f"sfn_role_{str(uuid4())[0:6]}"
        sfn_role = iam.create_role(
            RoleName=role_name,
            AssumeRolePolicyDocument=json.dumps(sfn_role_policy),
            Path="/",
        )["Role"]["Arn"]

        client = boto3.client("stepfunctions", region_name="us-east-1")

        name = f"sfn_name_{str(uuid4())[0:6]}"
        response = client.create_state_machine(
            name=name, definition=simple_definition, roleArn=sfn_role, publish=True
        )
        arn = response["stateMachineArn"]
        version_arn1 = response["stateMachineVersionArn"]

        # Use the initial version to describe the state machine
        version1 = client.describe_state_machine(stateMachineArn=version_arn1)
        assert version1["loggingConfiguration"] == {"level": "OFF"}

        # Update the state machine
        update = client.update_state_machine(
            stateMachineArn=arn,
            loggingConfiguration={"level": "ALL"},
            publish=True,
        )
        version_arn2 = update["stateMachineVersionArn"]
        assert version_arn1 != version_arn2

        # Assert that we can retrieve the latest configuration, either by the regular ARN or by the version ARN
        latest = client.describe_state_machine(stateMachineArn=arn)
        assert latest["loggingConfiguration"] == {"level": "ALL"}
        version2 = client.describe_state_machine(stateMachineArn=version_arn2)
        assert version2["loggingConfiguration"] == {"level": "ALL"}

        # Assert that we can still describe the first version of the state machine
        version1 = client.describe_state_machine(stateMachineArn=version_arn1)
        assert version1["loggingConfiguration"] == {"level": "OFF"}


@aws_verified
@pytest.mark.aws_verified
def test_create_state_machine_with_version_description():
    iam = boto3.client("iam", region_name="us-east-1")
    role_name = f"sfn_role_{str(uuid4())[0:6]}"
    sfn_role = iam.create_role(
        RoleName=role_name,
        AssumeRolePolicyDocument=json.dumps(sfn_role_policy),
        Path="/",
    )["Role"]["Arn"]

    client = boto3.client("stepfunctions", region_name="us-east-1")

    try:
        name = f"sfn_name_{str(uuid4())[0:6]}"

        response = client.create_state_machine(
            name=name,
            definition=simple_definition,
            roleArn=sfn_role,
            versionDescription="first version",
            publish=True,
        )
        arn = response["stateMachineArn"]
        version_arn1 = response["stateMachineVersionArn"]

        # Use the initial version to describe the state machine
        sm = client.describe_state_machine(stateMachineArn=arn)
        assert "description" not in sm

        version = client.describe_state_machine(stateMachineArn=version_arn1)
        assert version["description"] == "first version"

        update = client.update_state_machine(
            stateMachineArn=arn,
            definition=simple_definition,
            versionDescription="second version",
            publish=True,
        )
        version_arn2 = update["stateMachineVersionArn"]

        version = client.describe_state_machine(stateMachineArn=version_arn2)
        assert version["description"] == "second version"
    finally:
        client.delete_state_machine(stateMachineArn=arn)
        iam.delete_role(RoleName=role_name)


@aws_verified
@pytest.mark.aws_verified
def test_create_unpublished_state_machine_with_version_description():
    iam = boto3.client("iam", region_name="us-east-1")
    role_name = f"sfn_role_{str(uuid4())[0:6]}"
    sfn_role = iam.create_role(
        RoleName=role_name,
        AssumeRolePolicyDocument=json.dumps(sfn_role_policy),
        Path="/",
    )["Role"]["Arn"]

    client = boto3.client("stepfunctions", region_name="us-east-1")

    try:
        name = f"sfn_name_{str(uuid4())[0:6]}"

        with pytest.raises(ClientError) as exc:
            client.create_state_machine(
                name=name,
                definition=simple_definition,
                roleArn=sfn_role,
                versionDescription="first version of statemachine",
            )
        err = exc.value.response["Error"]
        assert err["Code"] == "ValidationException"
        assert (
            err["Message"] == "Version description can only be set when publish is true"
        )

    finally:
        iam.delete_role(RoleName=role_name)



import re
from typing import List, Optional

from antlr4.tree.Tree import ParseTree, TerminalNodeImpl

from moto.stepfunctions.parser.asl.antlr.runtime.ASLIntrinsicLexer import (
    ASLIntrinsicLexer,
)
from moto.stepfunctions.parser.asl.antlr.runtime.ASLIntrinsicParser import (
    ASLIntrinsicParser,
)
from moto.stepfunctions.parser.asl.antlr.runtime.ASLIntrinsicParserVisitor import (
    ASLIntrinsicParserVisitor,
)
from moto.stepfunctions.parser.asl.antlt4utils.antlr4utils import Antlr4Utils
from moto.stepfunctions.parser.asl.component.common.query_language import (
    QueryLanguageMode,
)
from moto.stepfunctions.parser.asl.component.common.variable_sample import (
    VariableSample,
)
from moto.stepfunctions.parser.asl.component.component import Component
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument import (
    FunctionArgument,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_bool import (
    FunctionArgumentBool,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_context_path import (
    FunctionArgumentContextPath,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_float import (
    FunctionArgumentFloat,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_function import (
    FunctionArgumentFunction,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_int import (
    FunctionArgumentInt,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_json_path import (
    FunctionArgumentJsonPath,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_list import (
    FunctionArgumentList,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_string import (
    FunctionArgumentString,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_var import (
    FunctionArgumentVar,
)
from moto.stepfunctions.parser.asl.component.intrinsic.function.function import Function
from moto.stepfunctions.parser.asl.component.intrinsic.function.statesfunction.factory import (
    StatesFunctionFactory,
)
from moto.stepfunctions.parser.asl.component.intrinsic.function.statesfunction.states_function import (
    StatesFunction,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.state_function_name_types import (
    StatesFunctionNameType,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.states_function_name import (
    StatesFunctionName,
)


class Preprocessor(ASLIntrinsicParserVisitor):
    @staticmethod
    def _replace_escaped_characters(match):
        escaped_char = match.group(1)
        if escaped_char.isalpha():
            replacements = {"n": "\n", "t": "\t", "r": "\r"}
            return replacements.get(escaped_char, escaped_char)
        elif escaped_char == '"':
            return '"'
        else:
            return match.group(0)

    @staticmethod
    def _text_of_str(parse_tree: ParseTree) -> str:
        pt = Antlr4Utils.is_production(parse_tree) or Antlr4Utils.is_terminal(
            parse_tree
        )
        inner_str = pt.getText()
        inner_str = inner_str[1:-1]
        inner_str = re.sub(
            r"\\(.)", Preprocessor._replace_escaped_characters, inner_str
        )
        return inner_str

    def visitFunc_arg_int(
        self, ctx: ASLIntrinsicParser.Func_arg_intContext
    ) -> FunctionArgumentInt:
        integer = int(ctx.INT().getText())
        return FunctionArgumentInt(integer=integer)

    def visitFunc_arg_float(
        self, ctx: ASLIntrinsicParser.Func_arg_floatContext
    ) -> FunctionArgumentFloat:
        number = float(ctx.INT().getText())
        return FunctionArgumentFloat(number=number)

    def visitFunc_arg_string(
        self, ctx: ASLIntrinsicParser.Func_arg_stringContext
    ) -> FunctionArgumentString:
        text: str = self._text_of_str(ctx.STRING())
        return FunctionArgumentString(string=text)

    def visitFunc_arg_context_path(
        self, ctx: ASLIntrinsicParser.Func_arg_context_pathContext
    ) -> FunctionArgumentContextPath:
        json_path: str = ctx.CONTEXT_PATH_STRING().getText()[1:]
        return FunctionArgumentContextPath(json_path=json_path)

    def visitFunc_arg_json_path(
        self, ctx: ASLIntrinsicParser.Func_arg_json_pathContext
    ) -> FunctionArgumentJsonPath:
        json_path: str = ctx.JSON_PATH_STRING().getText()
        return FunctionArgumentJsonPath(json_path=json_path)

    def visitFunc_arg_var(
        self, ctx: ASLIntrinsicParser.Func_arg_varContext
    ) -> FunctionArgumentVar:
        expression: str = ctx.STRING_VARIABLE().getText()
        variable_sample = VariableSample(
            query_language_mode=QueryLanguageMode.JSONPath, expression=expression
        )
        return FunctionArgumentVar(variable_sample=variable_sample)

    def visitFunc_arg_func_decl(
        self, ctx: ASLIntrinsicParser.Func_arg_func_declContext
    ) -> FunctionArgumentFunction:
        function: Function = self.visit(ctx.states_func_decl())
        return FunctionArgumentFunction(function=function)

    def visitFunc_arg_bool(
        self, ctx: ASLIntrinsicParser.Func_arg_boolContext
    ) -> FunctionArgumentBool:
        bool_term: TerminalNodeImpl = ctx.children[0]
        bool_term_rule: int = bool_term.getSymbol().type
        bool_val: bool = bool_term_rule == ASLIntrinsicLexer.TRUE
        return FunctionArgumentBool(boolean=bool_val)

    def visitFunc_arg_list(
        self, ctx: ASLIntrinsicParser.Func_arg_listContext
    ) -> FunctionArgumentList:
        arg_list: List[FunctionArgument] = list()
        for child in ctx.children:
            cmp: Optional[Component] = self.visit(child)
            if isinstance(cmp, FunctionArgument):
                arg_list.append(cmp)
        return FunctionArgumentList(arg_list=arg_list)

    def visitState_fun_name(
        self, ctx: ASLIntrinsicParser.State_fun_nameContext
    ) -> StatesFunctionName:
        tok_typ: int = ctx.children[0].symbol.type
        name_typ = StatesFunctionNameType(tok_typ)
        return StatesFunctionName(function_type=name_typ)

    def visitStates_func_decl(
        self, ctx: ASLIntrinsicParser.States_func_declContext
    ) -> StatesFunction:
        func_name: StatesFunctionName = self.visit(ctx.state_fun_name())
        arg_list: FunctionArgumentList = self.visit(ctx.func_arg_list())
        func: StatesFunction = StatesFunctionFactory.from_name(
            func_name=func_name, arg_list=arg_list
        )
        return func

    def visitFunc_decl(self, ctx: ASLIntrinsicParser.Func_declContext) -> Function:
        return self.visit(ctx.children[0])



"""KafkaBackend class with methods for supported APIs."""

import uuid
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.utils import get_partition

from ..utilities.tagging_service import TaggingService


class FakeKafkaCluster(BaseModel):
    def __init__(
        self,
        cluster_name: str,
        account_id: str,
        region_name: str,
        cluster_type: str,
        tags: Optional[Dict[str, str]] = None,
        broker_node_group_info: Optional[Dict[str, Any]] = None,
        kafka_version: Optional[str] = None,
        number_of_broker_nodes: Optional[int] = None,
        configuration_info: Optional[Dict[str, Any]] = None,
        serverless_config: Optional[Dict[str, Any]] = None,
        encryption_info: Optional[Dict[str, Any]] = None,
        enhanced_monitoring: str = "DEFAULT",
        open_monitoring: Optional[Dict[str, Any]] = None,
        logging_info: Optional[Dict[str, Any]] = None,
        storage_mode: str = "LOCAL",
        current_version: str = "1.0",
        client_authentication: Optional[Dict[str, Any]] = None,
        state: str = "CREATING",
        active_operation_arn: Optional[str] = None,
        zookeeper_connect_string: Optional[str] = None,
        zookeeper_connect_string_tls: Optional[str] = None,
    ):
        # General attributes
        self.cluster_id = str(uuid.uuid4())
        self.cluster_name = cluster_name
        self.account_id = account_id
        self.region_name = region_name
        self.cluster_type = cluster_type
        self.tags = tags or {}
        self.state = state
        self.creation_time = datetime.now().isoformat()
        self.current_version = current_version
        self.active_operation_arn = active_operation_arn
        self.arn = self._generate_arn()

        # Attributes specific to PROVISIONED clusters
        self.broker_node_group_info = broker_node_group_info
        self.kafka_version = kafka_version
        self.number_of_broker_nodes = number_of_broker_nodes
        self.configuration_info = configuration_info
        self.encryption_info = encryption_info
        self.enhanced_monitoring = enhanced_monitoring
        self.open_monitoring = open_monitoring
        self.logging_info = logging_info
        self.storage_mode = storage_mode
        self.client_authentication = client_authentication
        self.zookeeper_connect_string = zookeeper_connect_string
        self.zookeeper_connect_string_tls = zookeeper_connect_string_tls

        # Attributes specific to SERVERLESS clusters
        self.serverless_config = serverless_config

    def _generate_arn(self) -> str:
        resource_type = (
            "cluster" if self.cluster_type == "PROVISIONED" else "serverless-cluster"
        )
        partition = get_partition(self.region_name)
        return f"arn:{partition}:kafka:{self.region_name}:{self.account_id}:{resource_type}/{self.cluster_id}"


class KafkaBackend(BaseBackend):
    """Implementation of Kafka APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.clusters: Dict[str, FakeKafkaCluster] = {}
        self.tagger = TaggingService()

    def create_cluster_v2(
        self,
        cluster_name: str,
        tags: Optional[Dict[str, str]],
        provisioned: Optional[Dict[str, Any]],
        serverless: Optional[Dict[str, Any]],
    ) -> Tuple[str, str, str, str]:
        if provisioned:
            cluster_type = "PROVISIONED"
            broker_node_group_info = provisioned.get("brokerNodeGroupInfo")
            kafka_version = provisioned.get("kafkaVersion", "default-kafka-version")
            number_of_broker_nodes = int(provisioned.get("numberOfBrokerNodes", 1))
            storage_mode = provisioned.get("storageMode", "LOCAL")
            serverless_config = None
        elif serverless:
            cluster_type = "SERVERLESS"
            broker_node_group_info = None
            kafka_version = None
            number_of_broker_nodes = None
            storage_mode = None
            serverless_config = serverless

        new_cluster = FakeKafkaCluster(
            cluster_name=cluster_name,
            account_id=self.account_id,
            region_name=self.region_name,
            cluster_type=cluster_type,
            broker_node_group_info=broker_node_group_info,
            kafka_version=kafka_version,
            number_of_broker_nodes=number_of_broker_nodes,
            serverless_config=serverless_config,
            tags=tags,
            state="CREATING",
            storage_mode=storage_mode if storage_mode else "LOCAL",
            current_version="1.0",
        )

        self.clusters[new_cluster.arn] = new_cluster

        if tags:
            self.tag_resource(new_cluster.arn, tags)

        return (
            new_cluster.arn,
            new_cluster.cluster_name,
            new_cluster.state,
            new_cluster.cluster_type,
        )

    def describe_cluster_v2(self, cluster_arn: str) -> Dict[str, Any]:
        cluster = self.clusters[cluster_arn]

        cluster_info: Dict[str, Any] = {
            "activeOperationArn": "arn:aws:kafka:region:account-id:operation/active-operation",
            "clusterArn": cluster.arn,
            "clusterName": cluster.cluster_name,
            "clusterType": cluster.cluster_type,
            "creationTime": cluster.creation_time,
            "currentVersion": cluster.current_version,
            "state": cluster.state,
            "stateInfo": {
                "code": "string",
                "message": "Cluster state details.",
            },
            "tags": self.list_tags_for_resource(cluster.arn),
        }

        if cluster.cluster_type == "PROVISIONED":
            cluster_info.update(
                {
                    "provisioned": {
                        "brokerNodeGroupInfo": cluster.broker_node_group_info or {},
                        "clientAuthentication": cluster.client_authentication or {},
                        "currentBrokerSoftwareInfo": {
                            "configurationArn": (cluster.configuration_info or {}).get(
                                "arn", "string"
                            ),
                            "configurationRevision": (
                                cluster.configuration_info or {}
                            ).get("revision", 1),
                            "kafkaVersion": cluster.kafka_version,
                        },
                        "encryptionInfo": cluster.encryption_info or {},
                        "enhancedMonitoring": cluster.enhanced_monitoring,
                        "openMonitoring": cluster.open_monitoring or {},
                        "loggingInfo": cluster.logging_info or {},
                        "numberOfBrokerNodes": cluster.number_of_broker_nodes or 0,
                        "zookeeperConnectString": cluster.zookeeper_connect_string
                        or "zookeeper.example.com:2181",
                        "zookeeperConnectStringTls": cluster.zookeeper_connect_string_tls
                        or "zookeeper.example.com:2181",
                        "storageMode": cluster.storage_mode,
                        "customerActionStatus": "NONE",
                    }
                }
            )

        elif cluster.cluster_type == "SERVERLESS":
            cluster_info.update(
                {
                    "serverless": {
                        "vpcConfigs": cluster.serverless_config.get("vpcConfigs", [])
                        if cluster.serverless_config
                        else [],
                        "clientAuthentication": cluster.serverless_config.get(
                            "clientAuthentication", {}
                        )
                        if cluster.serverless_config
                        else {},
                    }
                }
            )

        return cluster_info

    def list_clusters_v2(
        self,
        cluster_name_filter: Optional[str],
        cluster_type_filter: Optional[str],
        max_results: Optional[int],
        next_token: Optional[str],
    ) -> Tuple[List[Dict[str, Any]], Optional[str]]:
        cluster_info_list = []
        for cluster_arn in self.clusters.keys():
            cluster_info = self.describe_cluster_v2(cluster_arn)
            cluster_info_list.append(cluster_info)

        return cluster_info_list, None

    def create_cluster(
        self,
        broker_node_group_info: Dict[str, Any],
        client_authentication: Optional[Dict[str, Any]],
        cluster_name: str,
        configuration_info: Optional[Dict[str, Any]] = None,
        encryption_info: Optional[Dict[str, Any]] = None,
        enhanced_monitoring: str = "DEFAULT",
        open_monitoring: Optional[Dict[str, Any]] = None,
        kafka_version: str = "2.8.1",
        logging_info: Optional[Dict[str, Any]] = None,
        number_of_broker_nodes: int = 1,
        tags: Optional[Dict[str, str]] = None,
        storage_mode: str = "LOCAL",
    ) -> Tuple[str, str, str]:
        new_cluster = FakeKafkaCluster(
            cluster_name=cluster_name,
            account_id=self.account_id,
            region_name=self.region_name,
            cluster_type="PROVISIONED",
            broker_node_group_info=broker_node_group_info,
            client_authentication=client_authentication,
            kafka_version=kafka_version,
            number_of_broker_nodes=number_of_broker_nodes,
            configuration_info=configuration_info,
            encryption_info=encryption_info,
            enhanced_monitoring=enhanced_monitoring,
            open_monitoring=open_monitoring,
            logging_info=logging_info,
            storage_mode=storage_mode,
        )

        self.clusters[new_cluster.arn] = new_cluster

        if tags:
            self.tag_resource(new_cluster.arn, tags)

        return new_cluster.arn, new_cluster.cluster_name, new_cluster.state

    def describe_cluster(self, cluster_arn: str) -> Dict[str, Any]:
        cluster = self.clusters[cluster_arn]

        return {
            "activeOperationArn": "arn:aws:kafka:region:account-id:operation/active-operation",
            "brokerNodeGroupInfo": cluster.broker_node_group_info or {},
            "clientAuthentication": cluster.client_authentication or {},
            "clusterArn": cluster.arn,
            "clusterName": cluster.cluster_name,
            "creationTime": cluster.creation_time,
            "currentBrokerSoftwareInfo": {
                "configurationArn": (cluster.configuration_info or {}).get(
                    "arn", "string"
                ),
                "configurationRevision": (cluster.configuration_info or {}).get(
                    "revision", 1
                ),
                "kafkaVersion": cluster.kafka_version,
            },
            "currentVersion": cluster.current_version,
            "encryptionInfo": cluster.encryption_info or {},
            "enhancedMonitoring": cluster.enhanced_monitoring,
            "openMonitoring": cluster.open_monitoring or {},
            "loggingInfo": cluster.logging_info or {},
            "numberOfBrokerNodes": cluster.number_of_broker_nodes or 0,
            "state": cluster.state,
            "stateInfo": {
                "code": "string",
                "message": "Cluster state details.",
            },
            "tags": self.list_tags_for_resource(cluster.arn),
            "zookeeperConnectString": cluster.zookeeper_connect_string
            or "zookeeper.example.com:2181",
            "zookeeperConnectStringTls": cluster.zookeeper_connect_string_tls
            or "zookeeper.example.com:2181",
            "storageMode": cluster.storage_mode,
            "customerActionStatus": "NONE",
        }

    def list_clusters(
        self,
        cluster_name_filter: Optional[str],
        max_results: Optional[int],
        next_token: Optional[str],
    ) -> List[Dict[str, Any]]:
        cluster_info_list = [
            {
                "clusterArn": cluster.arn,
                "clusterName": cluster.cluster_name,
                "state": cluster.state,
                "creationTime": cluster.creation_time,
                "clusterType": cluster.cluster_type,
            }
            for cluster_arn, cluster in self.clusters.items()
        ]

        return cluster_info_list

    def delete_cluster(self, cluster_arn: str, current_version: str) -> Tuple[str, str]:
        cluster = self.clusters.pop(cluster_arn)
        return cluster_arn, cluster.state

    def list_tags_for_resource(self, resource_arn: str) -> Dict[str, str]:
        return self.tagger.get_tag_dict_for_resource(resource_arn)

    def tag_resource(self, resource_arn: str, tags: Dict[str, str]) -> None:
        tags_list = [{"Key": k, "Value": v} for k, v in tags.items()]
        self.tagger.tag_resource(resource_arn, tags_list)

    def untag_resource(self, resource_arn: str, tag_keys: List[str]) -> None:
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)


kafka_backends = BackendDict(KafkaBackend, "kafka")



# LocalStack Resource Provider Scaffolding v2
from __future__ import annotations

from pathlib import Path
from typing import List, Optional, TypedDict

import localstack.services.cloudformation.provider_utils as util
from localstack.services.cloudformation.resource_provider import (
    OperationStatus,
    ProgressEvent,
    ResourceProvider,
    ResourceRequest,
)


class StepFunctionsActivityProperties(TypedDict):
    Name: Optional[str]
    Arn: Optional[str]
    Tags: Optional[List[TagsEntry]]


class TagsEntry(TypedDict):
    Key: Optional[str]
    Value: Optional[str]


REPEATED_INVOCATION = "repeated_invocation"


class StepFunctionsActivityProvider(ResourceProvider[StepFunctionsActivityProperties]):
    TYPE = "AWS::StepFunctions::Activity"  # Autogenerated. Don't change
    SCHEMA = util.get_schema_path(Path(__file__))  # Autogenerated. Don't change

    def create(
        self,
        request: ResourceRequest[StepFunctionsActivityProperties],
    ) -> ProgressEvent[StepFunctionsActivityProperties]:
        """
        Create a new resource.

        Primary identifier fields:
          - /properties/Arn

        Required properties:
          - Name

        Create-only properties:
          - /properties/Name

        Read-only properties:
          - /properties/Arn

        IAM permissions required:
          - states:CreateActivity

        """
        model = request.desired_state
        step_functions = request.aws_client_factory.stepfunctions
        if not model.get("Tags"):
            response = step_functions.create_activity(name=model["Name"])
        else:
            response = step_functions.create_activity(
                name=model["Name"], tags=model["Tags"]
            )
        model["Arn"] = response["activityArn"]

        return ProgressEvent(
            status=OperationStatus.SUCCESS,
            resource_model=model,
            custom_context=request.custom_context,
        )

    def read(
        self,
        request: ResourceRequest[StepFunctionsActivityProperties],
    ) -> ProgressEvent[StepFunctionsActivityProperties]:
        """
        Fetch resource information

        IAM permissions required:
          - states:DescribeActivity
          - states:ListTagsForResource
        """
        raise NotImplementedError

    def delete(
        self,
        request: ResourceRequest[StepFunctionsActivityProperties],
    ) -> ProgressEvent[StepFunctionsActivityProperties]:
        """
        Delete a resource

        IAM permissions required:
          - states:DeleteActivity
        """
        model = request.desired_state
        step_functions = request.aws_client_factory.stepfunctions

        step_functions.delete_activity(activityArn=model["Arn"])

        return ProgressEvent(
            status=OperationStatus.SUCCESS,
            resource_model=model,
            custom_context=request.custom_context,
        )

    def update(
        self,
        request: ResourceRequest[StepFunctionsActivityProperties],
    ) -> ProgressEvent[StepFunctionsActivityProperties]:
        """
        Update a resource

        IAM permissions required:
          - states:ListTagsForResource
          - states:TagResource
          - states:UntagResource
        """
        raise NotImplementedError



import random
import uuid
from unittest import SkipTest

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws, settings

SAMPLE_DOMAIN_NAME = "example.com"
SAMPLE_NAME_SERVERS = ["10.0.0.6", "10.0.0.7"]


@mock_aws
def test_dhcp_options_associate():
    """associate dhcp option"""
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")

    dhcp_options = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": [SAMPLE_DOMAIN_NAME]},
            {"Key": "domain-name-servers", "Values": SAMPLE_NAME_SERVERS},
        ]
    )
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    client.associate_dhcp_options(DhcpOptionsId=dhcp_options.id, VpcId=vpc.id)
    #
    vpc.reload()
    assert vpc.dhcp_options_id == dhcp_options.id


@mock_aws
def test_dhcp_options_associate_invalid_dhcp_id():
    """associate dhcp option bad dhcp options id"""
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    with pytest.raises(ClientError) as ex:
        client.associate_dhcp_options(DhcpOptionsId="foo", VpcId=vpc.id)
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidDhcpOptionID.NotFound"


@mock_aws
def test_dhcp_options_associate_invalid_vpc_id():
    """associate dhcp option invalid vpc id"""
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")
    dhcp_options = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": [SAMPLE_DOMAIN_NAME]},
            {"Key": "domain-name-servers", "Values": SAMPLE_NAME_SERVERS},
        ]
    )

    with pytest.raises(ClientError) as ex:
        client.associate_dhcp_options(DhcpOptionsId=dhcp_options.id, VpcId="foo")
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidVpcID.NotFound"


@mock_aws
def test_dhcp_options_disassociation():
    """Ensure that VPCs can be set to the 'default' DHCP options set for disassociation."""
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")
    dhcp_options = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": [SAMPLE_DOMAIN_NAME]},
            {"Key": "domain-name-servers", "Values": SAMPLE_NAME_SERVERS},
        ]
    )
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    # Ensure newly created VPCs without any DHCP options can be disassociated
    client.associate_dhcp_options(DhcpOptionsId="default", VpcId=vpc.id)
    vpc.reload()
    assert vpc.dhcp_options_id == "default"

    client.associate_dhcp_options(DhcpOptionsId=dhcp_options.id, VpcId=vpc.id)
    vpc.reload()
    assert vpc.dhcp_options_id == dhcp_options.id

    # Ensure VPCs with already associated DHCP options can be disassociated
    client.associate_dhcp_options(DhcpOptionsId="default", VpcId=vpc.id)
    vpc.reload()
    assert vpc.dhcp_options_id == "default"


@mock_aws
def test_dhcp_options_delete_with_vpc():
    """Test deletion of dhcp options with vpc"""
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")
    dhcp_options = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": [SAMPLE_DOMAIN_NAME]},
            {"Key": "domain-name-servers", "Values": SAMPLE_NAME_SERVERS},
        ]
    )
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    client.associate_dhcp_options(DhcpOptionsId=dhcp_options.id, VpcId=vpc.id)

    with pytest.raises(ClientError) as ex:
        client.delete_dhcp_options(DhcpOptionsId=dhcp_options.id)
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "DependencyViolation"

    vpc.delete()

    with pytest.raises(ClientError) as ex:
        client.describe_dhcp_options(DhcpOptionsIds=[dhcp_options.id])
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidDhcpOptionID.NotFound"


@mock_aws
def test_create_dhcp_options():
    """Create most basic dhcp option"""
    ec2 = boto3.resource("ec2", region_name="us-west-1")

    dhcp_options = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": [SAMPLE_DOMAIN_NAME]},
            {"Key": "domain-name-servers", "Values": SAMPLE_NAME_SERVERS},
        ]
    )
    config = dhcp_options.dhcp_configurations
    assert len(config) == 2
    assert {
        "Key": "domain-name-servers",
        "Values": [{"Value": ip} for ip in SAMPLE_NAME_SERVERS],
    } in config

    assert {"Key": "domain-name", "Values": [{"Value": SAMPLE_DOMAIN_NAME}]} in config


@mock_aws
def test_create_dhcp_options_invalid_options():
    """Create invalid dhcp options"""
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    servers = ["f", "f", "f", "f", "f"]

    with pytest.raises(ClientError) as ex:
        ec2.create_dhcp_options(
            DhcpConfigurations=[{"Key": "ntp-servers", "Values": servers}]
        )
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidParameterValue"

    with pytest.raises(ClientError) as ex:
        ec2.create_dhcp_options(
            DhcpConfigurations=[{"Key": "netbios-node-type", "Values": ["0"]}]
        )
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidParameterValue"


@mock_aws
def test_describe_dhcp_options():
    """Test dhcp options lookup by id"""
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")

    dhcp_options = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": [SAMPLE_DOMAIN_NAME]},
            {"Key": "domain-name-servers", "Values": SAMPLE_NAME_SERVERS},
        ]
    )
    all_options = client.describe_dhcp_options(DhcpOptionsIds=[dhcp_options.id])[
        "DhcpOptions"
    ]
    assert len(all_options) == 1

    all_options = client.describe_dhcp_options()["DhcpOptions"]
    assert len(all_options) >= 1, "Should have recently created DHCP option"
    recently_created = [
        o for o in all_options if o["DhcpOptionsId"] == dhcp_options.id
    ][0]
    assert recently_created["DhcpOptionsId"] == dhcp_options.id
    config = recently_created["DhcpConfigurations"]
    assert len(config) == 2
    assert {
        "Key": "domain-name-servers",
        "Values": [{"Value": ip} for ip in SAMPLE_NAME_SERVERS],
    } in config
    assert {"Key": "domain-name", "Values": [{"Value": SAMPLE_DOMAIN_NAME}]} in config


@mock_aws
def test_describe_dhcp_options_invalid_id():
    """get error on invalid dhcp_option_id lookup"""
    client = boto3.client("ec2", region_name="us-west-1")

    with pytest.raises(ClientError) as ex:
        client.describe_dhcp_options(DhcpOptionsIds=["1"])
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidDhcpOptionID.NotFound"


@mock_aws
def test_delete_dhcp_options():
    """delete dhcp option"""
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")

    dhcp_option = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": [SAMPLE_DOMAIN_NAME]},
            {"Key": "domain-name-servers", "Values": SAMPLE_NAME_SERVERS},
        ]
    )

    client.delete_dhcp_options(DhcpOptionsId=dhcp_option.id)

    with pytest.raises(ClientError) as ex:
        client.describe_dhcp_options(DhcpOptionsIds=[dhcp_option.id])
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidDhcpOptionID.NotFound"


@mock_aws
def test_delete_dhcp_options_invalid_id():
    client = boto3.client("ec2", region_name="us-west-1")

    with pytest.raises(ClientError) as ex:
        client.delete_dhcp_options(DhcpOptionsId="dopt-abcd1234")
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidDhcpOptionID.NotFound"


@mock_aws
def test_delete_dhcp_options_malformed_id():
    client = boto3.client("ec2", region_name="us-west-1")

    with pytest.raises(ClientError) as ex:
        client.delete_dhcp_options(DhcpOptionsId="foo-abcd1234")
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidDhcpOptionsId.Malformed"


@mock_aws
def test_dhcp_tagging():
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")

    dhcp_option = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": [SAMPLE_DOMAIN_NAME]},
            {"Key": "domain-name-servers", "Values": SAMPLE_NAME_SERVERS},
        ]
    )

    tag_value = str(uuid.uuid4())
    dhcp_option.create_tags(Tags=[{"Key": "a tag", "Value": tag_value}])

    tag = client.describe_tags(
        Filters=[{"Name": "resource-id", "Values": [dhcp_option.id]}]
    )["Tags"][0]
    assert tag["ResourceId"] == dhcp_option.id
    assert tag["ResourceType"] == "dhcp-options"
    assert tag["Key"] == "a tag"
    assert tag["Value"] == tag_value

    # Refresh the DHCP options
    dhcp_option = client.describe_dhcp_options(DhcpOptionsIds=[dhcp_option.id])[
        "DhcpOptions"
    ][0]
    assert dhcp_option["Tags"] == [{"Key": "a tag", "Value": tag_value}]


@mock_aws
def test_dhcp_options_get_by_tag():
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")

    dhcp_tag_value = str(uuid.uuid4())

    dhcp1 = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": ["example.com"]},
            {"Key": "domain-name-servers", "Values": ["10.0.10.2"]},
        ]
    )
    dhcp1_tag_name = str(uuid.uuid4())
    dhcp1.create_tags(
        Tags=[
            {"Key": "Name", "Value": dhcp1_tag_name},
            {"Key": "test-tag", "Value": dhcp_tag_value},
        ]
    )

    dhcp2 = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": ["example.com"]},
            {"Key": "domain-name-servers", "Values": ["10.0.20.2"]},
        ]
    )
    dhcp2_tag_name = str(uuid.uuid4())
    dhcp2.create_tags(
        Tags=[
            {"Key": "Name", "Value": dhcp2_tag_name},
            {"Key": "test-tag", "Value": dhcp_tag_value},
        ]
    )

    dhcp_options_sets = client.describe_dhcp_options(
        Filters=[
            {"Name": "tag:Name", "Values": [dhcp1_tag_name]},
            {"Name": "tag:test-tag", "Values": [dhcp_tag_value]},
        ]
    )["DhcpOptions"]

    assert len(dhcp_options_sets) == 1
    config = dhcp_options_sets[0]["DhcpConfigurations"]
    assert len(config) == 2
    assert {"Key": "domain-name", "Values": [{"Value": "example.com"}]} in config
    assert {"Key": "domain-name-servers", "Values": [{"Value": "10.0.10.2"}]} in config
    tags = dhcp_options_sets[0]["Tags"]
    assert len(tags) == 2
    assert {"Key": "Name", "Value": dhcp1_tag_name} in tags
    assert {"Key": "test-tag", "Value": dhcp_tag_value} in tags

    dhcp_options_sets = client.describe_dhcp_options(
        Filters=[
            {"Name": "tag:Name", "Values": [dhcp2_tag_name]},
            {"Name": "tag:test-tag", "Values": [dhcp_tag_value]},
        ]
    )["DhcpOptions"]

    assert len(dhcp_options_sets) == 1
    config = dhcp_options_sets[0]["DhcpConfigurations"]
    assert len(config) == 2
    assert {"Key": "domain-name", "Values": [{"Value": "example.com"}]} in config
    assert {"Key": "domain-name-servers", "Values": [{"Value": "10.0.20.2"}]} in config
    tags = dhcp_options_sets[0]["Tags"]
    assert len(tags) == 2
    assert {"Key": "Name", "Value": dhcp2_tag_name} in tags
    assert {"Key": "test-tag", "Value": dhcp_tag_value} in tags

    dhcp_options_sets = client.describe_dhcp_options(
        Filters=[{"Name": "tag:test-tag", "Values": [dhcp_tag_value]}]
    )["DhcpOptions"]

    assert len(dhcp_options_sets) == 2


@mock_aws
def test_dhcp_options_get_by_id():
    ec2 = boto3.resource("ec2", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")

    dhcp1 = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": ["test1.com"]},
            {"Key": "domain-name-servers", "Values": ["10.0.10.2"]},
        ]
    )
    dhcp1.create_tags(Tags=[{"Key": "Name", "Value": "TestDhcpOptions1"}])
    dhcp1.create_tags(Tags=[{"Key": "test-tag", "Value": "test-value"}])

    dhcp2 = ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": ["test2.com"]},
            {"Key": "domain-name-servers", "Values": ["10.0.20.2"]},
        ]
    )
    dhcp1.create_tags(Tags=[{"Key": "Name", "Value": "TestDhcpOptions2"}])
    dhcp1.create_tags(Tags=[{"Key": "test-tag", "Value": "test-value"}])

    options = client.describe_dhcp_options()["DhcpOptions"]
    d_ids = [d["DhcpOptionsId"] for d in options]
    assert dhcp1.id in d_ids
    assert dhcp2.id in d_ids

    d = client.describe_dhcp_options(
        Filters=[{"Name": "dhcp-options-id", "Values": [dhcp1.id]}]
    )["DhcpOptions"]

    assert len(d) == 1
    assert d[0]["DhcpOptionsId"] == dhcp1.id

    d = client.describe_dhcp_options(
        Filters=[{"Name": "dhcp-options-id", "Values": [dhcp2.id]}]
    )["DhcpOptions"]

    assert len(d) == 1
    assert d[0]["DhcpOptionsId"] == dhcp2.id


@mock_aws
def test_dhcp_options_get_by_value_filter():
    ec2 = boto3.resource("ec2", region_name="us-west-1")

    random_server_1 = ".".join(map(str, (random.randint(0, 99) for _ in range(4))))
    random_server_2 = ".".join(map(str, (random.randint(0, 99) for _ in range(4))))
    random_server_3 = ".".join(map(str, (random.randint(0, 99) for _ in range(4))))

    ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": ["example.com"]},
            {"Key": "domain-name-servers", "Values": [random_server_1]},
        ]
    )

    ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": ["example.com"]},
            {"Key": "domain-name-servers", "Values": [random_server_2]},
        ]
    )

    ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": ["example.com"]},
            {"Key": "domain-name-servers", "Values": [random_server_3]},
        ]
    )

    filters = [{"Name": "value", "Values": [random_server_2]}]
    dhcp_options_sets = list(ec2.dhcp_options_sets.filter(Filters=filters))
    assert len(dhcp_options_sets) == 1


@mock_aws
def test_dhcp_options_get_by_key_filter():
    ec2 = boto3.resource("ec2", region_name="us-west-1")

    random_domain_name = str(uuid.uuid4())[0:6]

    ec2.create_dhcp_options(
        DhcpConfigurations=[{"Key": "domain-name", "Values": [random_domain_name]}]
    )

    ec2.create_dhcp_options(
        DhcpConfigurations=[{"Key": "domain-name", "Values": ["example.com"]}]
    )

    ec2.create_dhcp_options(
        DhcpConfigurations=[{"Key": "domain-name", "Values": [random_domain_name]}]
    )

    filters = [{"Name": "key", "Values": ["domain-name"]}]
    dhcp_options_sets = list(ec2.dhcp_options_sets.filter(Filters=filters))
    assert (
        len(dhcp_options_sets) >= 3
    ), "Should have at least 3 DHCP options just created"

    configs = []
    for d in dhcp_options_sets:
        configs.extend(d.dhcp_configurations)

    servers = []
    for config in configs:
        if config["Key"] == "domain-name":
            servers.extend(config["Values"])
    assert {"Value": random_domain_name} in servers
    assert {"Value": "example.com"} in servers


@mock_aws
def test_dhcp_options_get_by_invalid_filter():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Will throw a generic 500 in ServerMode")
    ec2 = boto3.resource("ec2", region_name="us-west-1")

    client = boto3.client("ec2", region_name="us-west-1")

    ec2.create_dhcp_options(
        DhcpConfigurations=[
            {"Key": "domain-name", "Values": [SAMPLE_DOMAIN_NAME]},
            {"Key": "domain-name-servers", "Values": SAMPLE_NAME_SERVERS},
        ]
    )

    filters = [{"Name": "invalid-filter", "Values": ["n/a"]}]
    with pytest.raises(NotImplementedError):
        client.describe_dhcp_options(Filters=filters)



from typing import Any, Dict, List, Union

from moto.core.common_models import BaseModel
from moto.moto_api._internal import mock_random as random
from moto.utilities.utils import get_partition


class QuicksightDataSet(BaseModel):
    def __init__(self, account_id: str, region: str, _id: str, name: str):
        self.arn = f"arn:{get_partition(region)}:quicksight:{region}:{account_id}:data-set/{_id}"
        self._id = _id
        self.name = name
        self.region = region
        self.account_id = account_id

    def to_json(self) -> Dict[str, Any]:
        return {
            "Arn": self.arn,
            "DataSetId": self._id,
            "IngestionArn": f"arn:{get_partition(self.region)}:quicksight:{self.region}:{self.account_id}:ingestion/tbd",
        }


class QuicksightIngestion(BaseModel):
    def __init__(
        self, account_id: str, region: str, data_set_id: str, ingestion_id: str
    ):
        self.arn = f"arn:{get_partition(region)}:quicksight:{region}:{account_id}:data-set/{data_set_id}/ingestions/{ingestion_id}"
        self.ingestion_id = ingestion_id

    def to_json(self) -> Dict[str, Any]:
        return {
            "Arn": self.arn,
            "IngestionId": self.ingestion_id,
            "IngestionStatus": "INITIALIZED",
        }


class QuicksightMembership(BaseModel):
    def __init__(self, account_id: str, region: str, group: str, user: str):
        self.group = group
        self.user = user
        self.arn = f"arn:{get_partition(region)}:quicksight:{region}:{account_id}:group/default/{group}/{user}"

    def to_json(self) -> Dict[str, str]:
        return {"Arn": self.arn, "MemberName": self.user}


class QuicksightGroup(BaseModel):
    def __init__(
        self,
        region: str,
        group_name: str,
        description: str,
        aws_account_id: str,
        namespace: str,
    ):
        self.arn = f"arn:{get_partition(region)}:quicksight:{region}:{aws_account_id}:group/default/{group_name}"
        self.group_name = group_name
        self.description = description
        self.aws_account_id = aws_account_id
        self.namespace = namespace
        self.region = region

        self.members: Dict[str, QuicksightMembership] = dict()

    def add_member(self, member_name: str) -> QuicksightMembership:
        membership = QuicksightMembership(
            self.aws_account_id, self.region, self.group_name, member_name
        )
        self.members[member_name] = membership
        return membership

    def delete_member(self, user_name: str) -> None:
        self.members.pop(user_name, None)

    def get_member(self, user_name: str) -> Union[QuicksightMembership, None]:
        return self.members.get(user_name, None)

    def list_members(self) -> List[QuicksightMembership]:
        return list(self.members.values())

    def to_json(self) -> Dict[str, Any]:
        return {
            "Arn": self.arn,
            "GroupName": self.group_name,
            "Description": self.description,
            "PrincipalId": self.aws_account_id,
            "Namespace": self.namespace,
        }


class QuicksightUser(BaseModel):
    def __init__(
        self,
        account_id: str,
        region: str,
        email: str,
        identity_type: str,
        username: str,
        user_role: str,
    ):
        self.arn = f"arn:{get_partition(region)}:quicksight:{region}:{account_id}:user/default/{username}"
        self.email = email
        self.identity_type = identity_type
        self.username = username
        self.user_role = user_role
        self.active = False
        self.principal_id = random.get_random_hex(10)

    def to_json(self) -> Dict[str, Any]:
        return {
            "Arn": self.arn,
            "Email": self.email,
            "IdentityType": self.identity_type,
            "Role": self.user_role,
            "UserName": self.username,
            "Active": self.active,
            "PrincipalId": self.principal_id,
        }



"""FSxBackend class with methods for supported APIs."""

from typing import Any, Dict, List, Optional, Tuple
from uuid import uuid4

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.paginator import paginate

from .utils import FileSystemType

PAGINATION_MODEL = {
    "describe_file_systems": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 2147483647,
        "unique_attribute": "resource_arn",
    }
}


class FileSystem(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        file_system_type: str,
        storage_capacity: int,
        storage_type: str,
        subnet_ids: List[str],
        security_group_ids: List[str],
        tags: Optional[List[Dict[str, str]]],
        kms_key_id: Optional[str],
        windows_configuration: Optional[Dict[str, Any]],
        lustre_configuration: Optional[Dict[str, Any]],
        ontap_configuration: Optional[Dict[str, Any]],
        open_zfs_configuration: Optional[Dict[str, Any]],
    ) -> None:
        self.file_system_id = f"fs-{uuid4().hex[:8]}"
        self.file_system_type = file_system_type
        if self.file_system_type not in FileSystemType.list_values():
            raise ValueError(f"Invalid FileSystemType: {self.file_system_type}")
        self.storage_capacity = storage_capacity
        self.storage_type = storage_type
        self.subnet_ids = subnet_ids
        self.security_group_ids = security_group_ids
        self.dns_name = f"{self.file_system_id}.fsx.{region_name}.amazonaws.com"
        self.kms_key_id = kms_key_id
        self.resource_arn = (
            f"arn:aws:fsx:{region_name}:{account_id}:file-system/{self.file_system_id}"
        )
        self.tags = tags or []
        self.windows_configuration = windows_configuration
        self.lustre_configuration = lustre_configuration
        self.ontap_configuration = ontap_configuration
        self.open_zfs_configuration = open_zfs_configuration

    def to_dict(self) -> Dict[str, Any]:
        dct = {
            "FileSystemId": self.file_system_id,
            "FileSystemType": self.file_system_type,
            "StorageCapacity": self.storage_capacity,
            "StorageType": self.storage_type,
            "SubnetIds": self.subnet_ids,
            "SecurityGroupIds": self.security_group_ids,
            "Tags": self.tags,
            "DNSName": self.dns_name,
            "KmsKeyId": self.kms_key_id,
            "ResourceARN": self.resource_arn,
            "WindowsConfiguration": self.windows_configuration,
            "LustreConfiguration": self.lustre_configuration,
            "OntapConfiguration": self.ontap_configuration,
            "OpenZFSConfiguration": self.open_zfs_configuration,
        }
        return {k: v for k, v in dct.items() if v}


class FSxBackend(BaseBackend):
    """Implementation of FSx APIs."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.file_systems: Dict[str, FileSystem] = {}

    def create_file_system(
        self,
        client_request_token: str,
        file_system_type: str,
        storage_capacity: int,
        storage_type: str,
        subnet_ids: List[str],
        security_group_ids: List[str],
        tags: Optional[List[Dict[str, str]]],
        kms_key_id: Optional[str],
        windows_configuration: Optional[Dict[str, Any]],
        lustre_configuration: Optional[Dict[str, Any]],
        ontap_configuration: Optional[Dict[str, Any]],
        file_system_type_version: Optional[str],
        open_zfs_configuration: Optional[Dict[str, Any]],
    ) -> FileSystem:
        file_system = FileSystem(
            account_id=self.account_id,
            region_name=self.region_name,
            file_system_type=file_system_type,
            storage_capacity=storage_capacity,
            storage_type=storage_type,
            subnet_ids=subnet_ids,
            security_group_ids=security_group_ids,
            tags=tags,
            kms_key_id=kms_key_id,
            windows_configuration=windows_configuration,
            ontap_configuration=ontap_configuration,
            open_zfs_configuration=open_zfs_configuration,
            lustre_configuration=lustre_configuration,
        )

        file_system_id = file_system.file_system_id

        self.file_systems[file_system_id] = file_system
        return file_system

    @paginate(pagination_model=PAGINATION_MODEL)
    def describe_file_systems(self, file_system_ids: List[str]) -> List[FileSystem]:
        file_systems = []
        if not file_system_ids:
            file_systems = list(self.file_systems.values())
        else:
            for id in file_system_ids:
                if id in self.file_systems:
                    file_systems.append(self.file_systems[id])
        return file_systems

    def delete_file_system(
        self,
        file_system_id: str,
        client_request_token: str,
        windows_configuration: Optional[Dict[str, Any]],
        lustre_configuration: Optional[Dict[str, Any]],
        open_zfs_configuration: Optional[Dict[str, Any]],
    ) -> Tuple[
        str,
        str,
        Optional[Dict[str, Any]],
        Optional[Dict[str, Any]],
        Optional[Dict[str, Any]],
    ]:
        response_template = {"FinalBackUpId": "", "FinalBackUpTags": []}

        file_system_type = self.file_systems[file_system_id].file_system_type

        lifecycle = "DELETING"
        self.file_systems.pop(file_system_id)

        windows_response = None
        lustre_response = None
        open_zfs_response = None

        if file_system_type == "WINDOWS":
            windows_response = response_template
        elif file_system_type == "LUSTRE":
            lustre_response = response_template
        elif file_system_type == "OPEN_ZFS":
            open_zfs_response = response_template

        return (
            file_system_id,
            lifecycle,
            windows_response,
            lustre_response,
            open_zfs_response,
        )

    def tag_resource(self, resource_arn: str, tags: List[Dict[str, str]]) -> None:
        resource = self._get_resource_from_arn(resource_arn)
        resource.tags.extend(tags)

    def _get_resource_from_arn(self, arn: str) -> FileSystem:
        target_resource, target_name = arn.split(":")[-1].split("/")
        try:
            return self.file_systems[target_name]
        except KeyError:
            message = f"Could not find {target_resource} with name {target_name}"
            raise ValueError(message)

    def untag_resource(self, resource_arn: str, tag_keys: List[str]) -> None:
        resource = self._get_resource_from_arn(resource_arn)
        if tag_keys:
            resource.tags = [tag for tag in resource.tags if tag["Key"] not in tag_keys]


fsx_backends = BackendDict(FSxBackend, "fsx")



from __future__ import annotations

from typing import Optional

from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.flow.start_at import StartAt
from moto.stepfunctions.parser.asl.component.common.query_language import QueryLanguage
from moto.stepfunctions.parser.asl.component.program.states import States
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.distributed_iteration_component import (
    DistributedIterationComponent,
    DistributedIterationComponentEvalInput,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.distributed_item_processor_worker import (
    DistributedItemProcessorWorker,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.processor_config import (
    ProcessorConfig,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.parse.typed_props import TypedProps


class DistributedItemProcessorEvalInput(DistributedIterationComponentEvalInput):
    pass


class DistributedItemProcessor(DistributedIterationComponent):
    _eval_input: Optional[DistributedItemProcessorEvalInput]

    @classmethod
    def from_props(cls, props: TypedProps) -> DistributedItemProcessor:
        item_processor = cls(
            query_language=props.get(QueryLanguage) or QueryLanguage(),
            start_at=props.get(
                typ=StartAt,
                raise_on_missing=ValueError(
                    f"Missing StartAt declaration in props '{props}'."
                ),
            ),
            states=props.get(
                typ=States,
                raise_on_missing=ValueError(
                    f"Missing States declaration in props '{props}'."
                ),
            ),
            comment=props.get(Comment),
            processor_config=props.get(ProcessorConfig) or ProcessorConfig(),
        )
        return item_processor

    def _create_worker(self, env: Environment) -> DistributedItemProcessorWorker:
        return DistributedItemProcessorWorker(
            work_name=self._eval_input.state_name,
            job_pool=self._job_pool,
            env=env,
            item_reader=self._eval_input.item_reader,
            parameters=self._eval_input.parameters,
            item_selector=self._eval_input.item_selector,
            map_run_record=self._map_run_record,
        )



import json

import boto3

from moto import mock_aws

template_fs_simple = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Resources": {
        "FileSystemResource": {"Type": "AWS::EFS::FileSystem", "Properties": {}},
    },
}


template_complete = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Resources": {
        "MountTargetVPC": {
            "Type": "AWS::EC2::VPC",
            "Properties": {"CidrBlock": "172.31.0.0/16"},
        },
        "MountTargetSubnetOne": {
            "Type": "AWS::EC2::Subnet",
            "Properties": {
                "CidrBlock": "172.31.1.0/24",
                "VpcId": {"Ref": "MountTargetVPC"},
                "AvailabilityZone": "us-east-1a",
            },
        },
        "MountTargetSubnetTwo": {
            "Type": "AWS::EC2::Subnet",
            "Properties": {
                "CidrBlock": "172.31.2.0/24",
                "VpcId": {"Ref": "MountTargetVPC"},
                "AvailabilityZone": "us-east-1b",
            },
        },
        "MountTargetSubnetThree": {
            "Type": "AWS::EC2::Subnet",
            "Properties": {
                "CidrBlock": "172.31.3.0/24",
                "VpcId": {"Ref": "MountTargetVPC"},
                "AvailabilityZone": "us-east-1c",
            },
        },
        "FileSystemResource": {
            "Type": "AWS::EFS::FileSystem",
            "Properties": {
                "PerformanceMode": "maxIO",
                "LifecyclePolicies": [
                    {"TransitionToIA": "AFTER_30_DAYS"},
                    {"TransitionToPrimaryStorageClass": "AFTER_1_ACCESS"},
                ],
                "Encrypted": True,
                "FileSystemTags": [{"Key": "Name", "Value": "TestFileSystem"}],
                "FileSystemPolicy": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Action": ["elasticfilesystem:ClientMount"],
                            "Principal": {
                                "AWS": "arn:aws:iam::111122223333:role/EfsReadOnly"
                            },
                        }
                    ],
                },
                "BackupPolicy": {"Status": "ENABLED"},
                "KmsKeyId": {"Fn::GetAtt": ["key", "Arn"]},
            },
        },
        "key": {
            "Type": "AWS::KMS::Key",
            "Properties": {
                "KeyPolicy": {
                    "Version": "2012-10-17",
                    "Id": "key-default-1",
                    "Statement": [
                        {
                            "Sid": "Allow administration of the key",
                            "Effect": "Allow",
                            "Principal": {
                                "AWS": {
                                    "Fn::Join": [
                                        "",
                                        [
                                            "arn:aws:iam::",
                                            {"Ref": "AWS::AccountId"},
                                            ":root",
                                        ],
                                    ]
                                }
                            },
                            "Action": ["kms:*"],
                            "Resource": "*",
                        }
                    ],
                }
            },
        },
        "MountTargetResource1": {
            "Type": "AWS::EFS::MountTarget",
            "Properties": {
                "FileSystemId": {"Ref": "FileSystemResource"},
                "SubnetId": {"Ref": "MountTargetSubnetOne"},
                "SecurityGroups": [
                    {"Fn::GetAtt": ["MountTargetVPC", "DefaultSecurityGroup"]}
                ],
            },
        },
        "MountTargetResource2": {
            "Type": "AWS::EFS::MountTarget",
            "Properties": {
                "FileSystemId": {"Ref": "FileSystemResource"},
                "SubnetId": {"Ref": "MountTargetSubnetTwo"},
                "SecurityGroups": [
                    {"Fn::GetAtt": ["MountTargetVPC", "DefaultSecurityGroup"]}
                ],
            },
        },
        "MountTargetResource3": {
            "Type": "AWS::EFS::MountTarget",
            "Properties": {
                "FileSystemId": {"Ref": "FileSystemResource"},
                "SubnetId": {"Ref": "MountTargetSubnetThree"},
                "SecurityGroups": [
                    {"Fn::GetAtt": ["MountTargetVPC", "DefaultSecurityGroup"]}
                ],
            },
        },
        "AccessPointResource": {
            "Type": "AWS::EFS::AccessPoint",
            "Properties": {
                "FileSystemId": {"Ref": "FileSystemResource"},
                "PosixUser": {
                    "Uid": "13234",
                    "Gid": "1322",
                    "SecondaryGids": ["1344", "1452"],
                },
                "RootDirectory": {
                    "CreationInfo": {
                        "OwnerGid": "708798",
                        "OwnerUid": "7987987",
                        "Permissions": "0755",
                    },
                    "Path": "/testcfn/abc",
                },
            },
        },
    },
}


@mock_aws
def test_simple_template():
    region = "us-east-1"
    cf = boto3.client("cloudformation", region_name=region)
    cf.create_stack(StackName="teststack", TemplateBody=json.dumps(template_fs_simple))

    efs = boto3.client("efs", region)
    fs = efs.describe_file_systems()["FileSystems"][0]
    assert fs["PerformanceMode"] == "generalPurpose"
    assert fs["Encrypted"] is False
    assert fs["ThroughputMode"] == "bursting"


@mock_aws
def test_full_template():
    region = "us-east-1"
    cf = boto3.client("cloudformation", region_name=region)
    cf.create_stack(StackName="teststack", TemplateBody=json.dumps(template_complete))

    efs = boto3.client("efs", region)
    fs = efs.describe_file_systems()["FileSystems"][0]
    fs_id = fs["FileSystemId"]
    assert fs["Name"] == "TestFileSystem"
    assert fs["KmsKeyId"]

    lc = efs.describe_lifecycle_configuration(FileSystemId=fs_id)["LifecyclePolicies"]
    assert {"TransitionToIA": "AFTER_30_DAYS"} in lc
    assert {"TransitionToPrimaryStorageClass": "AFTER_1_ACCESS"} in lc

    aps = efs.describe_access_points()["AccessPoints"][0]
    assert aps["FileSystemId"] == fs_id

    cf.delete_stack(StackName="teststack")

    assert efs.describe_file_systems()["FileSystems"] == []
    assert efs.describe_access_points()["AccessPoints"] == []



from enum import Enum

PAGINATION_MODEL = {
    "describe_cache_clusters": {
        "input_token": "marker",
        "limit_key": "max_records",
        "limit_default": 100,
        "unique_attribute": "cache_cluster_id",
    },
}
VALID_AUTH_MODE_KEYS = ["Type", "Passwords"]
VALID_ENGINE_TYPES = ["redis", "valkey"]


class AuthenticationTypes(str, Enum):
    NOPASSWORD = "no-password-required"
    PASSWORD = "password"
    IAM = "iam"



"""securityhub base URL and path."""

from .responses import SecurityHubResponse

url_bases = [
    r"https?://securityhub\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/findings$": SecurityHubResponse.dispatch,
    "{0}/findings/import$": SecurityHubResponse.dispatch,
}



"""Unit tests for sagemaker-supported APIs."""

import datetime

import boto3

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_domain():
    client = boto3.client("sagemaker", region_name="us-east-2")
    resp = client.create_domain(
        DomainName="testdomain",
        AuthMode="IAM",
        DefaultUserSettings={
            "ExecutionRole": f"arn:aws:iam::{ACCOUNT_ID}:role/iamrole-domain-execution",
            "SecurityGroups": ["sg-12345678901234567"],
            "SharingSettings": {
                "NotebookOutputOption": "Allowed",
                "S3OutputPath": "s3://output",
            },
        },
        SubnetIds=["subnet-12345678901234567"],
        VpcId="VpcId",
        AppNetworkAccessType="PublicInternetOnly",
        Tags=[{"Key": "string", "Value": "string"}],
    )

    assert (
        resp["DomainArn"]
        == f"arn:aws:sagemaker:us-east-2:{ACCOUNT_ID}:domain/d-testdomain"
    )
    assert resp["Url"] == "testdomain.us-east-2.sagemaker.test.com"


@mock_aws
def test_describe_domain():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    client.create_domain(
        DomainName="testdomain",
        AuthMode="IAM",
        DefaultUserSettings={
            "ExecutionRole": f"arn:aws:iam::{ACCOUNT_ID}:role/iamrole-domain-execution",
            "SecurityGroups": ["sg-12345678901234567"],
            "SharingSettings": {
                "NotebookOutputOption": "Allowed",
                "S3OutputPath": "s3://output",
            },
        },
        SubnetIds=["subnet-12345678901234567"],
        VpcId="VpcId",
        AppNetworkAccessType="PublicInternetOnly",
        HomeEfsFileSystemKmsKeyId="testHomeEfsFileSystemKmsKeyId",
        KmsKeyId="testKmsKeyId",
        DefaultSpaceSettings={
            "ExecutionRole": f"arn:aws:iam::{ACCOUNT_ID}:role/iamrole-space-execution",
            "SecurityGroups": [
                "sg-12345678901234564",
            ],
            "JupyterServerAppSettings": {
                "DefaultResourceSpec": {
                    "SageMakerImageArn": "SageMakerImageArn",
                    "SageMakerImageVersionArn": "SageMakerImageVersionArn",
                    "SageMakerImageVersionAlias": "SageMakerImageVersionAlias",
                    "InstanceType": "system",
                    "LifecycleConfigArn": "LifecycleConfigArn",
                },
                "LifecycleConfigArns": [
                    "LifecycleConfigArns",
                ],
                "CodeRepositories": [
                    {"RepositoryUrl": "RepositoryUrl"},
                ],
            },
        },
        Tags=[{"Key": "string", "Value": "string"}],
    )
    resp = client.describe_domain(DomainId="d-testdomain")
    assert (
        resp["DomainArn"]
        == f"arn:aws:sagemaker:ap-southeast-1:{ACCOUNT_ID}:domain/d-testdomain"
    )
    assert resp["DomainId"] == "d-testdomain"
    assert resp["DomainName"] == "testdomain"
    assert resp["HomeEfsFileSystemId"] == "testdomain-efs-id"
    assert resp["SingleSignOnManagedApplicationInstanceId"] == "testdomain-sso-id"
    assert (
        resp["SingleSignOnApplicationArn"]
        == "arn:aws:sagemaker::123456789012:sso/application/testdomain/apl-testdomain"
    )
    assert resp["Status"] == "InService"
    assert isinstance(resp["CreationTime"], datetime.datetime)
    assert isinstance(resp["LastModifiedTime"], datetime.datetime)
    assert resp["FailureReason"] == ""
    assert resp["SecurityGroupIdForDomainBoundary"] == "sg-testdomain"
    assert resp["AuthMode"] == "IAM"
    assert resp["DefaultUserSettings"] == {
        "ExecutionRole": f"arn:aws:iam::{ACCOUNT_ID}:role/iamrole-domain-execution",
        "SecurityGroups": ["sg-12345678901234567"],
        "SharingSettings": {
            "NotebookOutputOption": "Allowed",
            "S3OutputPath": "s3://output",
        },
    }


@mock_aws
def test_describe_domain_defaults():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    resp = client.create_domain(
        DomainName="testdomain",
        AuthMode="IAM",
        DefaultUserSettings={},
        SubnetIds=["subnet-12345678901234567"],
        VpcId="VpcId",
    )
    resp = client.describe_domain(DomainId="d-testdomain")
    assert (
        resp["DomainArn"]
        == f"arn:aws:sagemaker:ap-southeast-1:{ACCOUNT_ID}:domain/d-testdomain"
    )
    assert resp["DomainId"] == "d-testdomain"
    assert resp["DomainName"] == "testdomain"
    assert resp["HomeEfsFileSystemId"] == "testdomain-efs-id"
    assert resp["SingleSignOnManagedApplicationInstanceId"] == "testdomain-sso-id"
    assert (
        resp["SingleSignOnApplicationArn"]
        == "arn:aws:sagemaker::123456789012:sso/application/testdomain/apl-testdomain"
    )
    assert resp["Status"] == "InService"
    assert isinstance(resp["CreationTime"], datetime.datetime)
    assert isinstance(resp["LastModifiedTime"], datetime.datetime)
    assert resp["FailureReason"] == ""
    assert resp["SecurityGroupIdForDomainBoundary"] == "sg-testdomain"
    assert resp["AuthMode"] == "IAM"
    assert resp["DefaultUserSettings"] == {}


@mock_aws
def test_list_domains():
    client = boto3.client("sagemaker", region_name="us-east-2")
    client.create_domain(
        DomainName="testdomain",
        AuthMode="IAM",
        DefaultUserSettings={},
        SubnetIds=["subnet-12345678901234567"],
        VpcId="VpcId",
    )
    client.create_domain(
        DomainName="testdomain2",
        AuthMode="IAM",
        DefaultUserSettings={},
        SubnetIds=["subnet-12345678901234567"],
        VpcId="VpcId",
    )
    resp = client.list_domains()
    assert len(resp["Domains"]) == 2
    assert (
        resp["Domains"][0]["DomainArn"]
        == "arn:aws:sagemaker:us-east-2:123456789012:domain/d-testdomain"
    )
    assert resp["Domains"][0]["DomainId"] == "d-testdomain"
    assert resp["Domains"][0]["DomainName"] == "testdomain"
    assert resp["Domains"][0]["Status"] == "InService"
    assert isinstance(resp["Domains"][0]["CreationTime"], datetime.datetime)
    assert isinstance(resp["Domains"][0]["LastModifiedTime"], datetime.datetime)
    assert resp["Domains"][0]["Url"] == "testdomain.us-east-2.sagemaker.test.com"
    assert (
        resp["Domains"][1]["DomainArn"]
        == "arn:aws:sagemaker:us-east-2:123456789012:domain/d-testdomain2"
    )
    assert resp["Domains"][1]["DomainId"] == "d-testdomain2"
    assert resp["Domains"][1]["DomainName"] == "testdomain2"
    assert resp["Domains"][1]["Status"] == "InService"
    assert isinstance(resp["Domains"][1]["CreationTime"], datetime.datetime)
    assert isinstance(resp["Domains"][1]["LastModifiedTime"], datetime.datetime)
    assert resp["Domains"][1]["Url"] == "testdomain2.us-east-2.sagemaker.test.com"


@mock_aws
def test_delete_domain():
    client = boto3.client("sagemaker", region_name="us-east-2")
    client.create_domain(
        DomainName="testdomain",
        AuthMode="IAM",
        DefaultUserSettings={},
        SubnetIds=["subnet-12345678901234567"],
        VpcId="VpcId",
    )
    client.create_domain(
        DomainName="testdomain2",
        AuthMode="IAM",
        DefaultUserSettings={},
        SubnetIds=["subnet-12345678901234567"],
        VpcId="VpcId",
    )
    client.delete_domain(DomainId="d-testdomain2")
    resp = client.list_domains()
    assert len(resp["Domains"]) == 1
    assert (
        resp["Domains"][0]["DomainArn"]
        == "arn:aws:sagemaker:us-east-2:123456789012:domain/d-testdomain"
    )
    assert resp["Domains"][0]["DomainId"] == "d-testdomain"
    assert resp["Domains"][0]["DomainName"] == "testdomain"


@mock_aws
def test_tag_domain():
    client = boto3.client("sagemaker", region_name="us-east-2")
    resp = client.create_domain(
        DomainName="testdomain",
        AuthMode="IAM",
        DefaultUserSettings={
            "ExecutionRole": f"arn:aws:iam::{ACCOUNT_ID}:role/iamrole-domain-execution",
            "SecurityGroups": ["sg-12345678901234567"],
            "SharingSettings": {
                "NotebookOutputOption": "Allowed",
                "S3OutputPath": "s3://output",
            },
        },
        SubnetIds=["subnet-12345678901234567"],
        VpcId="VpcId",
        AppNetworkAccessType="PublicInternetOnly",
        Tags=[{"Key": "testkey", "Value": "testvalue"}],
    )
    tags1 = client.list_tags(ResourceArn=resp["DomainArn"])
    assert tags1["Tags"] == [{"Key": "testkey", "Value": "testvalue"}]
    client.add_tags(
        ResourceArn=resp["DomainArn"], Tags=[{"Key": "testkey2", "Value": "testvalue2"}]
    )
    tags2 = client.list_tags(ResourceArn=resp["DomainArn"])
    assert tags2["Tags"] == [
        {"Key": "testkey", "Value": "testvalue"},
        {"Key": "testkey2", "Value": "testvalue2"},
    ]
    client.delete_tags(ResourceArn=resp["DomainArn"], TagKeys=["testkey"])
    tags3 = client.list_tags(ResourceArn=resp["DomainArn"])
    assert tags3["Tags"] == [{"Key": "testkey2", "Value": "testvalue2"}]



"""Exceptions raised by the qldb service."""

from moto.core.exceptions import JsonRESTError


class QLDBError(JsonRESTError):
    code = 400


class LedgerNotFoundException(QLDBError):
    def __init__(self, name: str) -> None:
        super().__init__(
            "LedgerNotFound",
            f"{name} does not match any existing ledger.",
        )


class LedgerNameAlreadyTakenException(QLDBError):
    def __init__(self, name: str) -> None:
        super().__init__(
            "LedgerNameAlreadyTaken",
            f"There is already a ledger with the name '{name}'",
        )


class ResourceNotFoundException(QLDBError):
    def __init__(self, resource_arn: str) -> None:
        super().__init__(
            "ResourceNotFound",
            f"{resource_arn} does not match the resource arn of any existing ledger",
        )



# mypy: disable-error-code="misc, var-annotated"

from __future__ import annotations

from collections import defaultdict
from datetime import datetime
from functools import lru_cache
from typing import Any, Dict, List, Mapping, MutableMapping, Optional, Tuple, Union

import xmltodict
from botocore import xform_name
from botocore.model import (
    ListShape,
    NoShapeFoundError,
    OperationModel,
    Shape,
    StructureShape,
)
from botocore.utils import parse_to_aware_datetime

from .utils import get_service_model

Serialized = MutableMapping[str, Any]


class ErrorShape(StructureShape):
    pass


class ShapeHelpersMixin:
    @staticmethod
    def get_serialized_name(shape: Shape, default_name: str) -> str:
        return shape.serialization.get("name", default_name)


class TimestampSerializer:
    TIMESTAMP_FORMAT_ISO8601 = "iso8601"
    TIMESTAMP_FORMAT_RFC822 = "rfc822"
    TIMESTAMP_FORMAT_UNIX = "unixtimestamp"

    ISO8601 = "%Y-%m-%dT%H:%M:%SZ"
    ISO8601_MICRO = "%Y-%m-%dT%H:%M:%S.%fZ"

    def __init__(self, default_format: str) -> None:
        self.default_format = default_format

    def serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        timestamp_format = shape.serialization.get(
            "timestampFormat", self.default_format
        )
        serialized_value = self._convert_timestamp_to_str(value, timestamp_format)
        serialized[key] = serialized_value

    def _timestamp_iso8601(self, value: datetime) -> str:
        timestamp_format = self.ISO8601
        if value.microsecond > 0:
            timestamp_format = self.ISO8601_MICRO
        return value.strftime(timestamp_format)

    def _convert_timestamp_to_str(
        self, value: Union[int, str, datetime], timestamp_format: str
    ) -> str:
        timestamp_format = timestamp_format.lower()
        converter = getattr(self, "_timestamp_%s" % timestamp_format)
        datetime_obj = parse_to_aware_datetime(value)  # type: ignore
        final_value = converter(datetime_obj)
        return final_value


class Serializer(ShapeHelpersMixin):  # , BaseSerializer):
    DEFAULT_RESPONSE_CODE = 200
    DEFAULT_ERROR_RESPONSE_CODE = 400
    # Clients can change this to a different MutableMapping
    # (i.e. OrderedDict) if they want.  This is used in the
    # compliance test to match the hash ordering used in the
    # tests.
    # NOTE: This is no longer necessary because dicts post 3.6 are ordered
    # https://stackoverflow.com/questions/39980323/are-dictionaries-ordered-in-python-3-6
    MAP_TYPE = dict
    DEFAULT_ENCODING = "utf-8"

    # From the spec, the default timestamp format if not specified is iso8601.
    DEFAULT_TIMESTAMP_FORMAT = TimestampSerializer.TIMESTAMP_FORMAT_ISO8601

    def __init__(
        self,
        operation_model: OperationModel,
        context: Optional[dict[str, Any]] = None,
        value_picker: Any = None,
        pretty_print: bool = False,
    ) -> None:
        self.operation_model = operation_model
        self.context = context or {"request_id": "request-id"}
        self.pretty_print = pretty_print
        if value_picker is None:
            value_picker = self._default_value_picker
        self._value_picker = value_picker
        self._timestamp_serializer = TimestampSerializer(self.DEFAULT_TIMESTAMP_FORMAT)

    def serialize_to_response(
        self,
        result: Any,
    ) -> Mapping[str, Any]:
        raise NotImplementedError("serialize_to_response")

    def _create_default_response(self) -> Serialized:
        # Creates a boilerplate default request dict that subclasses
        # can use as a starting point.
        serialized = {
            "status_code": self.DEFAULT_RESPONSE_CODE,
            "headers": {},
            # An empty body is represented as an empty string.
            "body": "",
        }
        return serialized

    # Some extra utility methods subclasses can use.

    @staticmethod
    def _is_error_result(result: object) -> bool:
        return isinstance(result, Exception)

    @staticmethod
    def _default_value_picker(obj: Any, key: str, _: Shape, default: Any = None) -> Any:
        if not hasattr(obj, "__getitem__"):
            return getattr(obj, key, default)

        try:
            return obj[key]
        except (KeyError, IndexError, TypeError, AttributeError):
            return getattr(obj, key, default)

    def _get_value(self, value: Any, key: str, shape: Shape) -> Any:
        return self._value_picker(value, key, shape)


class ResponseSerializer(Serializer):
    DEFAULT_TIMESTAMP_FORMAT = TimestampSerializer.TIMESTAMP_FORMAT_ISO8601

    CONTENT_TYPE = "text"

    def _encode_body(self, body: Any) -> str:
        raise NotImplementedError("_encode_body")

    @staticmethod
    def _get_error_shape_name(error: Exception) -> str:
        shape_name = getattr(error, "code", error.__class__.__name__)
        return shape_name

    def _get_error_shape(
        self, error: Exception, operation_model: OperationModel
    ) -> ErrorShape:
        shape_name = self._get_error_shape_name(error)
        try:
            # TODO: there is also an errors array in the operation model,
            # but I think it only includes the possible errors for that
            # operation.  Maybe we try that first, then try all shapes?
            shape = operation_model.service_model.shape_for(shape_name)
            # We convert to ErrorShape to keep mypy happy...
            shape = ErrorShape(
                shape_name,
                shape._shape_model,  # type: ignore
                shape._shape_resolver,  # type: ignore
            )
        except NoShapeFoundError:
            generic_error_model = {
                "exception": True,
                "type": "structure",
                "members": {},
                "error": {
                    "code": shape_name,
                },
            }
            shape = ErrorShape(shape_name, generic_error_model)
        return shape

    def _serialize_error(
        self,
        serialized: Serialized,
        error: Exception,
        operation_model: OperationModel,
        request_ctx: Mapping[str, Any],
    ) -> Serialized:
        shape = self._get_error_shape(error, operation_model)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        serialized["status_code"] = status_code
        message = getattr(error, "message", None) or str(error)
        error_wrapper, error_body = self._get_error_wrapper(
            operation_model, request_ctx
        )
        self._inject_error_metadata(error_body, error, shape, operation_model)
        if message:
            error_body["Message"] = message
        if shape is not None:
            self._serialize(error_body, error, shape, "")
        serialized["body"] = error_wrapper
        self._inject_error_headers(serialized["headers"], shape, operation_model)
        return serialized

    def _get_error_wrapper(
        self, operation_model: OperationModel, request_ctx: Mapping[str, Any]
    ) -> Tuple[Serialized, Serialized]:
        raise NotImplementedError("_get_error_wrapper")

    def _inject_error_metadata(
        self,
        serialized: Serialized,
        error: Exception,
        shape: ErrorShape,
        operation_model: OperationModel,
    ) -> None:
        raise NotImplementedError("_inject_error_metadata")

    def _inject_error_headers(
        self, serialized: Serialized, shape: ErrorShape, operation_model: OperationModel
    ) -> None:
        pass

    def _inject_response_metadata(
        self,
        serialized: Serialized,
        operation_model: OperationModel,
        request_ctx: Mapping[str, Any],
    ) -> None:
        raise NotImplementedError("_inject_response_metadata")

    def serialize_to_response(
        self,
        result: Any,
    ) -> Serialized:
        serialized = self._create_default_response()
        if self._is_error_result(result):
            serialized = self._serialize_error(
                serialized, result, self.operation_model, self.context
            )
        else:
            response_wrapper, result_wrapper = self._get_response_wrapper(
                self.operation_model, self.context
            )
            output_shape = self.operation_model.output_shape
            if output_shape is not None:
                self._serialize(result_wrapper, result, output_shape, "")

            root_key = list(response_wrapper.keys())[0]
            self._inject_response_metadata(
                response_wrapper[root_key], self.operation_model, self.context
            )

            serialized["body"] = response_wrapper

        serialized["body"] = self._encode_body(serialized["body"])

        serialized["headers"]["Content-Type"] = self.CONTENT_TYPE
        return serialized

    def _serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        method = getattr(
            self, "_serialize_type_%s" % shape.type_name, self._default_serialize
        )
        method(serialized, value, shape, key)

    def _serialize_type_structure(
        self, serialized: Serialized, value: Any, shape: StructureShape, key: str
    ) -> None:
        if key:
            new_serialized = self.MAP_TYPE()
            serialized[key] = new_serialized
            serialized = new_serialized
        for member_key, member_shape in shape.members.items():
            self._serialize_structure_member(
                serialized, value, member_shape, member_key
            )

    def _serialize_structure_member(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        member_value = self._get_value(value, key, shape)
        if member_value is not None:
            key_name = self.get_serialized_name(shape, key)
            self._serialize(serialized, member_value, shape, key_name)

    @staticmethod
    def _default_serialize(
        serialized: Serialized, value: Any, _: Shape, key: str
    ) -> None:
        serialized[key] = value

    def _serialize_type_timestamp(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        value_wrapper = {}
        value_key = "timestamp"
        self._timestamp_serializer.serialize(value_wrapper, value, shape, value_key)
        self._default_serialize(serialized, value_wrapper[value_key], shape, key)

    def _get_response_wrapper(
        self, operation_model: OperationModel, request_ctx: Mapping[str, Any]
    ) -> Tuple[Serialized, Serialized]:
        raise NotImplementedError("shouldn't get here...")


class BaseXMLSerializer(ResponseSerializer):
    @staticmethod
    def _serialize_namespace_attribute(
        serialized: Serialized, operation_model: OperationModel
    ) -> None:
        if "xmlNamespace" in operation_model.metadata:
            namespace = operation_model.metadata["xmlNamespace"]
            serialized["@xmlns"] = namespace

    def _get_error_wrapper(
        self, operation_model: OperationModel, request_ctx: Mapping[str, Any]
    ) -> Tuple[Serialized, Serialized]:
        serialized_error = self.MAP_TYPE()
        error_wrapper = {
            "ErrorResponse": {
                "Error": serialized_error,
                "RequestId": request_ctx.get("request_id"),
            }
        }
        self._serialize_namespace_attribute(
            error_wrapper["ErrorResponse"], operation_model
        )
        return error_wrapper, serialized_error

    def _get_response_wrapper(
        self, operation_model: OperationModel, request_ctx: Mapping[str, Any]
    ) -> Tuple[Serialized, Serialized]:
        serialized_result = self.MAP_TYPE()
        root_key = f"{operation_model.name}Response"
        response_wrapper = {root_key: {}}
        output_shape = operation_model.output_shape
        result_key = None
        if output_shape is not None:
            result_key = output_shape.serialization.get("resultWrapper")
        if result_key is not None:
            response_wrapper[root_key][result_key] = serialized_result
        else:
            response_wrapper[root_key] = serialized_result
        self._serialize_namespace_attribute(response_wrapper[root_key], operation_model)
        return response_wrapper, serialized_result

    def _inject_error_metadata(
        self,
        serialized: Serialized,
        error: Exception,
        shape: ErrorShape,
        operation_model: OperationModel,
    ) -> None:
        sender_fault = shape.metadata.get("error", {}).get("senderFault", True)
        serialized["Type"] = "Sender" if sender_fault else "Receiver"
        serialized["Code"] = shape.error_code
        message = getattr(error, "message", None) or str(error)
        if message:
            serialized["Message"] = message
        if shape is not None:
            self._serialize(serialized, error, shape, "")

    def _encode_body(self, body: Serialized) -> str:
        body_encoded = xmltodict.unparse(
            body,
            full_document=False,
            short_empty_elements=True,
            pretty=self.pretty_print,
        )
        return body_encoded

    def _serialize_type_list(
        self, serialized: Serialized, value: Any, shape: ListShape, key: str
    ) -> None:
        list_obj = []
        items_name = self.get_serialized_name(shape.member, "member")
        serialized[key] = {items_name: list_obj}
        for list_item in value:
            wrapper = {}
            self._serialize(wrapper, list_item, shape.member, "__current__")
            list_obj.append(wrapper["__current__"])
        if not list_obj:
            serialized[key] = ""


class QuerySerializer(BaseXMLSerializer):
    CONTENT_TYPE = "text/xml"

    def _inject_response_metadata(
        self,
        serialized: MutableMapping[str, Any],
        operation_model: OperationModel,
        request_ctx: Mapping[str, Any],
    ) -> None:
        serialized["ResponseMetadata"] = {"RequestId": request_ctx.get("request_id")}


SERIALIZERS = {
    "query": QuerySerializer,
}


class XFormedAttributeAccessMixin:
    """Mixin allowing access to "xformed" attributes:

    obj.DBInstanceIdentifier will retrieve the value of obj.db_instance_identifier

    """

    BOTOCORE_MODEL: Optional[str] = None

    _model_attribute_aliases: Dict[str, List[str]] = {}
    _xform_cache: Dict[str, str] = {}

    def __getattr__(self, name: str) -> Any:
        if name in self.model_attributes:
            return self.get_modeled_attribute(name)
        raise AttributeError(f"Attribute '{name}' not found!")

    def get_modeled_attribute(self, attr_name: str) -> Any:
        for attr_alias in self.model_attribute_aliases[attr_name]:
            try:
                return super().__getattribute__(attr_alias)
            except AttributeError:
                pass
        else:
            raise AttributeError

    @property
    def model_attributes(self) -> List[str]:
        return list(self.model_attribute_aliases.keys())

    @property
    def model_attribute_aliases(self) -> Dict[str, List[str]]:
        if not self._model_attribute_aliases:
            self._model_attribute_aliases = self.get_model_attributes_info()
        return self._model_attribute_aliases

    @classmethod
    @lru_cache()
    def get_model_attributes_info(cls) -> Dict[str, List[str]]:
        service_name = cls.__module__.split(".")[1]
        model_name = cls.BOTOCORE_MODEL or cls.__name__
        service_model = get_service_model(service_name)
        model_shape = service_model.shape_for(model_name)
        valid_attributes: Dict[str, List[str]] = defaultdict(list)
        for member_name, member_shape in model_shape.members.items():  # type: ignore[attr-defined]
            aliases = valid_attributes[member_name]
            if member_shape.type_name == "list":
                if member_name != member_shape.name:
                    xformed_name = cls._xform_name(member_shape.name)
                    aliases.append(xformed_name)
            xformed_member_name = cls._xform_name(member_name)
            aliases.append(xformed_member_name)
            if member_name.startswith(model_name):
                short_name = member_name[len(model_name) :]
                xformed_short_name = cls._xform_name(short_name)
                aliases.append(xformed_short_name)
        return valid_attributes

    @classmethod
    def _xform_name(cls, name: str) -> str:
        return xform_name(name, _xform_cache=cls._xform_cache)



"""Unit tests for securityhub-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID


@mock_aws
def test_get_findings():
    client = boto3.client("securityhub", region_name="us-east-1")

    test_finding = {
        "AwsAccountId": DEFAULT_ACCOUNT_ID,
        "CreatedAt": "2024-01-01T00:00:00.001Z",
        "UpdatedAt": "2024-01-01T00:00:00.000Z",
        "Description": "Test finding description",
        "GeneratorId": "test-generator",
        "Id": "test-finding-001",
        "ProductArn": f"arn:aws:securityhub:{client.meta.region_name}:{DEFAULT_ACCOUNT_ID}:product/{DEFAULT_ACCOUNT_ID}/default",
        "Resources": [{"Id": "test-resource", "Type": "AwsEc2Instance"}],
        "SchemaVersion": "2018-10-08",
        "Severity": {"Label": "HIGH"},
        "Title": "Test Finding",
        "Types": ["Software and Configuration Checks"],
    }

    import_response = client.batch_import_findings(Findings=[test_finding])
    assert import_response["SuccessCount"] == 1

    response = client.get_findings()

    assert "Findings" in response
    assert isinstance(response["Findings"], list)
    assert len(response["Findings"]) == 1
    finding = response["Findings"][0]
    assert finding["Id"] == "test-finding-001"
    assert finding["SchemaVersion"] == "2018-10-08"


@mock_aws
def test_batch_import_findings():
    client = boto3.client("securityhub", region_name="us-east-2")

    valid_finding = {
        "AwsAccountId": DEFAULT_ACCOUNT_ID,
        "CreatedAt": "2024-01-01T00:00:00.000Z",
        "UpdatedAt": "2024-01-01T00:00:00.000Z",
        "Description": "Test finding description",
        "GeneratorId": "test-generator",
        "Id": "test-finding-001",
        "ProductArn": f"arn:aws:securityhub:{client.meta.region_name}:{DEFAULT_ACCOUNT_ID}:product/{DEFAULT_ACCOUNT_ID}/default",
        "Resources": [{"Id": "test-resource", "Type": "AwsEc2Instance"}],
        "SchemaVersion": "2018-10-08",
        "Severity": {"Label": "HIGH"},
        "Title": "Test Finding",
        "Types": ["Software and Configuration Checks"],
    }

    response = client.batch_import_findings(Findings=[valid_finding])
    assert response["SuccessCount"] == 1
    assert response["FailedCount"] == 0
    assert response["FailedFindings"] == []

    invalid_finding = valid_finding.copy()
    invalid_finding["Id"] = "test-finding-002"
    invalid_finding["Severity"]["Label"] = "INVALID_LABEL"

    response = client.batch_import_findings(Findings=[invalid_finding])

    assert response["SuccessCount"] == 1
    assert response["FailedCount"] == 0
    assert len(response["FailedFindings"]) == 0


@mock_aws
def test_get_findings_invalid_parameters():
    client = boto3.client("securityhub", region_name="us-east-1")

    with pytest.raises(ClientError) as exc:
        client.get_findings(MaxResults=101)

    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInputException"
    assert "MaxResults must be a number between 1 and 100" in err["Message"]


@mock_aws
def test_batch_import_multiple_findings():
    client = boto3.client("securityhub", region_name="us-east-1")

    findings = [
        {
            "AwsAccountId": DEFAULT_ACCOUNT_ID,
            "CreatedAt": "2024-01-01T00:00:00.000Z",
            "UpdatedAt": "2024-01-01T00:00:00.000Z",
            "Description": f"Test finding description {i}",
            "GeneratorId": "test-generator",
            "Id": f"test-finding-{i:03d}",
            "ProductArn": f"arn:aws:securityhub:{client.meta.region_name}:{DEFAULT_ACCOUNT_ID}:product/{DEFAULT_ACCOUNT_ID}/default",
            "Resources": [{"Id": f"test-resource-{i}", "Type": "AwsEc2Instance"}],
            "SchemaVersion": "2018-10-08",
            "Severity": {"Label": "HIGH"},
            "Title": f"Test Finding {i}",
            "Types": ["Software and Configuration Checks"],
        }
        for i in range(1, 4)
    ]

    import_response = client.batch_import_findings(Findings=findings)
    assert import_response["SuccessCount"] == 3
    assert import_response["FailedCount"] == 0
    assert import_response["FailedFindings"] == []

    get_response = client.get_findings()
    assert "Findings" in get_response
    assert isinstance(get_response["Findings"], list)
    assert len(get_response["Findings"]) == 3

    imported_ids = {finding["Id"] for finding in get_response["Findings"]}
    expected_ids = {f"test-finding-{i:03d}" for i in range(1, 4)}
    assert imported_ids == expected_ids


@mock_aws
def test_get_findings_max_results():
    client = boto3.client("securityhub", region_name="us-east-1")

    findings = [
        {
            "AwsAccountId": DEFAULT_ACCOUNT_ID,
            "CreatedAt": "2024-01-01T00:00:00.000Z",
            "UpdatedAt": "2024-01-01T00:00:00.000Z",
            "Description": f"Test finding description {i}",
            "GeneratorId": "test-generator",
            "Id": f"test-finding-{i:03d}",
            "ProductArn": f"arn:aws:securityhub:{client.meta.region_name}:{DEFAULT_ACCOUNT_ID}:product/{DEFAULT_ACCOUNT_ID}/default",
            "Resources": [{"Id": f"test-resource-{i}", "Type": "AwsEc2Instance"}],
            "SchemaVersion": "2018-10-08",
            "Severity": {"Label": "HIGH"},
            "Title": f"Test Finding {i}",
            "Types": ["Software and Configuration Checks"],
        }
        for i in range(1, 4)
    ]

    import_response = client.batch_import_findings(Findings=findings)
    assert import_response["SuccessCount"] == 3

    get_response = client.get_findings(MaxResults=1)
    assert "Findings" in get_response
    assert isinstance(get_response["Findings"], list)
    assert len(get_response["Findings"]) == 1
    assert "NextToken" in get_response



from moto.stepfunctions.parser.asl.component.component import Component
from moto.stepfunctions.parser.asl.component.state.state import CommonStateField


class States(Component):
    def __init__(self):
        self.states: dict[str, CommonStateField] = dict()



import base64
import logging
from typing import Tuple

from moto.stepfunctions.parser.api import ValidationException
from moto.stepfunctions.parser.utils import to_bytes, to_str

LOG = logging.getLogger(__name__)


def get_next_page_token_from_arn(resource_arn: str) -> str:
    return to_str(base64.b64encode(to_bytes(resource_arn)))


_DEFAULT_SFN_MAX_RESULTS: int = 100


def normalise_max_results(max_results: int = 100) -> int:
    if not max_results:
        return _DEFAULT_SFN_MAX_RESULTS
    return max_results


def assert_pagination_parameters_valid(
    max_results: int,
    next_token: str,
    next_token_length_limit: int = 1024,
    max_results_upper_limit: int = 1000,
) -> Tuple[int, str]:
    validation_errors = []

    if isinstance(max_results, int) and max_results > max_results_upper_limit:
        validation_errors.append(
            f"Value '{max_results}' at 'maxResults' failed to satisfy constraint: "
            f"Member must have value less than or equal to {max_results_upper_limit}"
        )

    if isinstance(next_token, str) and len(next_token) > next_token_length_limit:
        validation_errors.append(
            f"Value '{next_token}' at 'nextToken' failed to satisfy constraint: "
            f"Member must have length less than or equal to {next_token_length_limit}"
        )

    if validation_errors:
        errors_message = "; ".join(validation_errors)
        message = f"{len(validation_errors)} validation {'errors' if len(validation_errors) > 1 else 'error'} detected: {errors_message}"
        raise ValidationException(message)



from typing import Any, Dict, List, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import unix_time
from moto.moto_api._internal import mock_random
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import get_partition

from .exceptions import AppNotFound, AppVersionNotFound, ResiliencyPolicyNotFound

PAGINATION_MODEL = {
    "list_apps": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "arn",
    },
    "list_resiliency_policies": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "arn",
    },
}


class AppComponent(BaseModel):
    def __init__(self, _id: str, name: str, _type: str):
        self.id = _id
        self.name = name
        self.type = _type

    def to_json(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "type": self.type,
        }


class App(BaseModel):
    def __init__(
        self,
        backend: "ResilienceHubBackend",
        assessment_schedule: str,
        description: str,
        event_subscriptions: List[Dict[str, Any]],
        name: str,
        permission_model: Dict[str, Any],
        policy_arn: str,
    ):
        self.backend = backend
        self.arn = f"arn:{get_partition(backend.region_name)}:resiliencehub:{backend.region_name}:{backend.account_id}:app/{mock_random.uuid4()}"
        self.assessment_schedule = assessment_schedule or "Disabled"
        self.compliance_status = "NotAssessed"
        self.description = description
        self.creation_time = unix_time()
        self.event_subscriptions = event_subscriptions
        self.name = name
        self.permission_model = permission_model
        self.policy_arn = policy_arn
        self.resilience_score = 0.0
        self.status = "Active"
        self.app_versions: List[AppVersion] = []

        app_version = AppVersion(app_arn=self.arn, version_name=None, identifier=0)
        self.app_versions.append(app_version)

    def get_version(self, version_name: str) -> "AppVersion":
        for v in self.app_versions:
            if v.app_version == version_name:
                return v
        raise AppVersionNotFound

    def to_json(self) -> Dict[str, Any]:
        resp = {
            "appArn": self.arn,
            "assessmentSchedule": self.assessment_schedule,
            "complianceStatus": self.compliance_status,
            "creationTime": self.creation_time,
            "driftStatus": "NotChecked",
            "name": self.name,
            "resilienceScore": self.resilience_score,
            "status": self.status,
            "tags": self.backend.list_tags_for_resource(self.arn),
        }
        if self.description is not None:
            resp["description"] = self.description
        if self.event_subscriptions:
            resp["eventSubscriptions"] = self.event_subscriptions
        if self.permission_model:
            resp["permissionModel"] = self.permission_model
        if self.policy_arn:
            resp["policyArn"] = self.policy_arn
        return resp


class Resource:
    def __init__(
        self,
        logical_resource_id: Dict[str, Any],
        physical_resource_id: str,
        resource_type: str,
        components: List[AppComponent],
    ):
        self.logical_resource_id = logical_resource_id
        self.physical_resource_id = physical_resource_id
        self.resource_type = resource_type
        self.components = components

    def to_json(self) -> Dict[str, Any]:
        return {
            "appComponents": [c.to_json() for c in self.components],
            "resourceType": self.resource_type,
            "logicalResourceId": self.logical_resource_id,
            "physicalResourceId": {"identifier": self.physical_resource_id},
            "resourceName": self.logical_resource_id["identifier"],
        }


class AppVersion(BaseModel):
    def __init__(self, app_arn: str, version_name: Optional[str], identifier: int):
        self.app_arn = app_arn
        self.eks_sources: List[Dict[str, Any]] = []
        self.source_arns: List[str] = []
        self.terraform_sources: List[Dict[str, str]] = []
        self.app_version = "release" if version_name else "draft"
        self.identifier = identifier
        self.creation_time = unix_time()
        self.version_name = version_name
        self.app_components: List[AppComponent] = []
        self.status = "Pending"
        self.resources: List[Resource] = []

    def to_json(self) -> Dict[str, Any]:
        resp = {
            "appVersion": self.app_version,
            "creationTime": self.creation_time,
            "identifier": self.identifier,
        }
        if self.version_name:
            resp["versionName"] = self.version_name
        return resp


class Policy(BaseModel):
    def __init__(
        self,
        backend: "ResilienceHubBackend",
        policy: Dict[str, Dict[str, int]],
        policy_name: str,
        data_location_constraint: str,
        policy_description: str,
        tier: str,
    ):
        self.arn = f"arn:{get_partition(backend.region_name)}:resiliencehub:{backend.region_name}:{backend.account_id}:resiliency-policy/{mock_random.uuid4()}"
        self.backend = backend
        self.data_location_constraint = data_location_constraint
        self.creation_time = unix_time()
        self.policy = policy
        self.policy_description = policy_description
        self.policy_name = policy_name
        self.tier = tier

    def to_json(self) -> Dict[str, Any]:
        resp = {
            "creationTime": self.creation_time,
            "policy": self.policy,
            "policyArn": self.arn,
            "policyName": self.policy_name,
            "tags": self.backend.list_tags_for_resource(self.arn),
            "tier": self.tier,
        }
        if self.data_location_constraint:
            resp["dataLocationConstraint"] = self.data_location_constraint
        if self.policy_description:
            resp["policyDescription"] = self.policy_description
        return resp


class ResilienceHubBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.apps: Dict[str, App] = dict()
        self.policies: Dict[str, Policy] = dict()
        self.tagger = TaggingService()

        self.app_assessments_queue: List[List[Dict[str, Any]]] = []
        self.app_assessments_results: Dict[str, List[Dict[str, Any]]] = {}

    def create_app(
        self,
        assessment_schedule: str,
        description: str,
        event_subscriptions: List[Dict[str, Any]],
        name: str,
        permission_model: Dict[str, Any],
        policy_arn: str,
        tags: Dict[str, str],
    ) -> App:
        """
        The ClientToken-parameter is not yet implemented
        """
        app = App(
            backend=self,
            assessment_schedule=assessment_schedule,
            description=description,
            event_subscriptions=event_subscriptions,
            name=name,
            permission_model=permission_model,
            policy_arn=policy_arn,
        )
        self.apps[app.arn] = app
        self.tag_resource(app.arn, tags)
        return app

    def create_resiliency_policy(
        self,
        data_location_constraint: str,
        policy: Dict[str, Any],
        policy_description: str,
        policy_name: str,
        tags: Dict[str, str],
        tier: str,
    ) -> Policy:
        """
        The ClientToken-parameter is not yet implemented
        """
        pol = Policy(
            backend=self,
            data_location_constraint=data_location_constraint,
            policy=policy,
            policy_description=policy_description,
            policy_name=policy_name,
            tier=tier,
        )
        self.policies[pol.arn] = pol
        self.tag_resource(pol.arn, tags)
        return pol

    @paginate(PAGINATION_MODEL)
    def list_apps(self, app_arn: str, name: str, reverse_order: bool) -> List[App]:
        """
        The FromAssessmentTime/ToAssessmentTime-parameters are not yet implemented
        """
        if name:
            app_summaries = [a for a in self.apps.values() if a.name == name]
        elif app_arn:
            app_summaries = [self.apps[app_arn]]
        else:
            app_summaries = list(self.apps.values())
        if reverse_order:
            app_summaries.reverse()
        return app_summaries

    def list_app_assessments(self, request_identifier: str) -> List[Dict[str, Any]]:
        """
        Moto will not actually execute any assessments, so this operation will return an empty list by default.
        You can use a dedicated API to override this, by configuring a queue of expected results.

        A request to `list_app_assessments` will take the first result from that queue, with subsequent calls with the same parameters returning the same result.

        Calling `list_app_assessments` with a different set of parameters will return the second result from that queue - and so on, or an empty list of the queue is empty.

        Configure this queue by making an HTTP request to `/moto-api/static/resilience-hub-assessments/response`. An example invocation looks like this:

        .. sourcecode:: python

            summary1 = {"appArn": "app_arn1", "appVersion": "some version", ...}
            summary2 = {"appArn": "app_arn2", ...}
            results = {"results": [[summary1, summary2], [summary2]], "region": "us-east-1"}
            resp = requests.post(
                "http://motoapi.amazonaws.com/moto-api/static/resilience-hub-assessments/response",
                json=results,
            )

            assert resp.status_code == 201

            client = boto3.client("lambda", region_name="us-east-1")
            # First result
            resp = client.list_app_assessments() # [summary1, summary2]
            # Second result
            resp = client.list_app_assessments(assessmentStatus="Pending") # [summary2]

        If you're using MotoServer, make sure to make this request to where MotoServer is running:

        .. sourcecode:: python

            http://localhost:5000/moto-api/static/resilience-hub-assessments/response

        """
        if request_identifier in self.app_assessments_results:
            return self.app_assessments_results[request_identifier]
        if self.app_assessments_queue:
            self.app_assessments_results[request_identifier] = (
                self.app_assessments_queue.pop(0)
            )
            return self.app_assessments_results[request_identifier]
        return []

    def describe_app(self, app_arn: str) -> App:
        if app_arn not in self.apps:
            raise AppNotFound(app_arn)
        return self.apps[app_arn]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_resiliency_policies(self, policy_name: str) -> List[Policy]:
        if policy_name:
            return [p for p in self.policies.values() if p.policy_name == policy_name]
        return list(self.policies.values())

    def describe_resiliency_policy(self, policy_arn: str) -> Policy:
        if policy_arn not in self.policies:
            raise ResiliencyPolicyNotFound(policy_arn)
        return self.policies[policy_arn]

    def tag_resource(self, resource_arn: str, tags: Dict[str, str]) -> None:
        self.tagger.tag_resource(
            resource_arn, TaggingService.convert_dict_to_tags_input(tags)
        )

    def untag_resource(self, resource_arn: str, tag_keys: List[str]) -> None:
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)

    def list_tags_for_resource(self, resource_arn: str) -> Dict[str, str]:
        return self.tagger.get_tag_dict_for_resource(resource_arn)

    def import_resources_to_draft_app_version(
        self,
        app_arn: str,
        eks_sources: List[Dict[str, Any]],
        source_arns: List[str],
        terraform_sources: List[Dict[str, str]],
    ) -> AppVersion:
        app = self.describe_app(app_arn)
        app_version = app.get_version("draft")

        app_version.eks_sources.extend(eks_sources)
        app_version.source_arns.extend(source_arns)
        app_version.terraform_sources.extend(terraform_sources)

        # Default AppComponent when importing data
        # AWS seems to create other components as well, based on the provided sources
        app_version.app_components.append(
            AppComponent(
                _id="appcommon",
                name="appcommon",
                _type="AWS::ResilienceHub::AppCommonAppComponent",
            )
        )
        return app_version

    def create_app_version_app_component(
        self, app_arn: str, name: str, _type: str
    ) -> AppComponent:
        app = self.describe_app(app_arn)
        app_version = app.get_version("draft")
        component = AppComponent(_id=name, name=name, _type=_type)
        app_version.app_components.append(component)
        return component

    def list_app_version_app_components(
        self, app_arn: str, app_version: str
    ) -> List[AppComponent]:
        app = self.describe_app(app_arn)
        return app.get_version(app_version).app_components

    def create_app_version_resource(
        self,
        app_arn: str,
        app_components: List[str],
        logical_resource_id: Dict[str, str],
        physical_resource_id: str,
        resource_type: str,
    ) -> Resource:
        app = self.describe_app(app_arn)
        app_version = app.get_version("draft")

        components = [c for c in app_version.app_components if c.id in app_components]

        resource = Resource(
            logical_resource_id=logical_resource_id,
            physical_resource_id=physical_resource_id,
            resource_type=resource_type,
            components=components,
        )
        app_version.resources.append(resource)
        return resource

    def list_app_version_resources(
        self, app_arn: str, app_version: str
    ) -> List[Resource]:
        app = self.describe_app(app_arn)
        return app.get_version(app_version).resources

    def list_app_versions(self, app_arn: str) -> List[AppVersion]:
        app = self.describe_app(app_arn)
        return app.app_versions

    def publish_app_version(self, app_arn: str, version_name: str) -> AppVersion:
        app = self.describe_app(app_arn)
        version = AppVersion(
            app_arn=app_arn, version_name=version_name, identifier=len(app.app_versions)
        )
        for old_version in app.app_versions:
            if old_version.app_version == "release":
                old_version.app_version = str(old_version.identifier)
        app.app_versions.append(version)
        return version


resiliencehub_backends = BackendDict(ResilienceHubBackend, "resiliencehub")
